# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import warnings

import h5py
import numpy as np
import pytest
import scipp as sc
from scipp.testing import assert_identical

import scippnexus as snx
from scippnexus import (
    NexusStructureError,
    NXdetector,
    NXentry,
    NXevent_data,
    NXlog,
    NXmonitor,
    NXroot,
)

# representative sample of UTF-8 test strings from
# https://www.w3.org/2001/06/utf-8-test/UTF-8-demo.html
UTF8_TEST_STRINGS = (
    "∮ E⋅da = Q,  n → ∞, ∑ f(i) = ∏ g(i), ∀x∈ℝ: ⌈x⌉ = −⌊−x⌋, α ∧ ¬β = ¬(¬α ∨ β)",  # noqa: RUF001
    "2H₂ + O₂ ⇌ 2H₂O, R = 4.7 kΩ, ⌀ 200 mm",
    "Σὲ γνωρίζω ἀπὸ τὴν κόψη",
)


@pytest.fixture
def h5root():
    """Yield h5py root group (file)"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        yield f


@pytest.fixture
def nxroot():
    """Yield NXroot containing a single NXentry named 'entry'"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        root = snx.Group(f, definitions=snx.base_definitions())
        root.create_class('entry', NXentry)
        yield root


def test_nxobject_root(nxroot) -> None:
    assert nxroot.nx_class == NXroot
    assert set(nxroot.keys()) == {'entry'}


def test_nxobject_create_class_creates_keys(nxroot) -> None:
    nxroot.create_class('log', NXlog)
    assert set(nxroot.keys()) == {'entry', 'log'}


def test_nxobject_create_class_with_string_nx_class(nxroot) -> None:
    nxroot.create_class('log', 'NXlog')
    assert set(nxroot.keys()) == {'entry', 'log'}


def test_nxobject_items(nxroot) -> None:
    items = nxroot.items()
    assert len(items) == 1
    name, entry = next(iter(items))
    assert name == 'entry'
    entry.create_class('monitor', NXmonitor)
    entry.create_class('log', NXlog)
    assert {k: v.nx_class for k, v in entry.items()} == {
        'log': NXlog,
        'monitor': NXmonitor,
    }


def test_nxobject_iter(nxroot) -> None:
    nxroot.create_class('entry2', NXentry)
    # With missing __iter__ this used to raise since Python just calls __getitem__
    # with an int range.
    list(nxroot)
    for _ in nxroot:
        pass


def test_nxobject_entry(nxroot) -> None:
    entry = nxroot['entry']
    assert entry.nx_class == NXentry
    entry.create_class('events_0', NXevent_data)
    entry.create_class('events_1', NXevent_data)
    entry.create_class('log', NXlog)
    assert set(entry.keys()) == {'events_0', 'events_1', 'log'}


def test_nx_class_can_be_bytes(h5root) -> None:
    log = h5root.create_group('log')
    attr = np.array(b'NXlog', dtype='|S5')
    log.attrs['NX_class'] = attr
    group = snx.Group(log, definitions=snx.base_definitions())
    assert group.nx_class == NXlog


def test_nxobject_event_data(nxroot) -> None:
    event_data = nxroot['entry'].create_class('events_0', NXevent_data)
    assert event_data.nx_class == NXevent_data


def test_nxobject_getting_item_that_does_not_exists_raises_KeyError(nxroot) -> None:
    with pytest.raises(KeyError):
        nxroot['abcde']


def test_nxobject_name_property_is_full_path(nxroot) -> None:
    nxroot.create_class('monitor', NXmonitor)
    nxroot['entry'].create_class('log', NXlog)
    nxroot['entry'].create_class('events_0', NXevent_data)
    assert nxroot.name == '/'
    assert nxroot['monitor'].name == '/monitor'
    assert nxroot['entry'].name == '/entry'
    assert nxroot['entry']['log'].name == '/entry/log'
    assert nxroot['entry']['events_0'].name == '/entry/events_0'


def test_nxobject_grandchild_can_be_accessed_using_path(nxroot) -> None:
    nxroot['entry'].create_class('log', NXlog)
    assert nxroot['entry/log'].name == '/entry/log'
    assert nxroot['/entry/log'].name == '/entry/log'


def test_nxobject_getitem_by_class(nxroot) -> None:
    nxroot.create_class('monitor', NXmonitor)
    nxroot['entry'].create_class('log', NXlog)
    nxroot['entry'].create_class('events_0', NXevent_data)
    nxroot['entry'].create_class('events_1', NXevent_data)
    assert list(nxroot[NXentry]) == ['entry']
    assert list(nxroot[NXmonitor]) == ['monitor']
    assert list(nxroot['entry'][NXmonitor]) == []  # not nested
    assert list(nxroot[NXlog]) == []  # nested
    assert list(nxroot['entry'][NXlog]) == ['log']
    assert set(nxroot['entry'][NXevent_data]) == {'events_0', 'events_1'}


def test_nxobject_getitem_by_class_get_fields(nxroot) -> None:
    nxroot['entry'].create_class('log', NXlog)
    nxroot['entry'].create_class('events_0', NXevent_data)
    nxroot['entry']['field1'] = sc.arange('event', 4.0, unit='ns')
    nxroot['entry']['field2'] = sc.arange('event', 2.0, unit='ns')
    assert list(nxroot[snx.Field]) == []
    assert set(nxroot['entry'][snx.Field]) == {'field1', 'field2'}


def test_nxobject_getitem_by_class_list(nxroot) -> None:
    nxroot['entry'].create_class('log', NXlog)
    nxroot['entry'].create_class('events_0', NXevent_data)
    nxroot['entry'].create_class('events_1', NXevent_data)
    nxroot['entry']['field1'] = sc.arange('event', 4.0, unit='ns')
    assert set(nxroot['entry'][[NXlog, NXevent_data]]) == {
        'log',
        'events_0',
        'events_1',
    }
    assert set(nxroot['entry'][[NXlog, snx.Field]]) == {'log', 'field1'}


def test_nxobject_dataset_items_are_returned_as_Field(nxroot) -> None:
    events = nxroot['entry'].create_class('events_0', NXevent_data)
    events['event_time_offset'] = sc.arange('event', 5)
    field = nxroot['entry/events_0/event_time_offset']
    assert isinstance(field, snx.Field)


def test_field_properties(nxroot) -> None:
    events = nxroot['entry'].create_class('events_0', NXevent_data)
    events['event_time_offset'] = sc.arange('event', 6, dtype='int64', unit='ns')
    field = nxroot['entry/events_0/event_time_offset']
    assert field.dtype == 'int64'
    assert field.name == '/entry/events_0/event_time_offset'
    assert field.shape == (6,)
    assert field.unit == sc.Unit('ns')


def test_field_dim_labels(nxroot) -> None:
    events = nxroot['entry'].create_class('events_0', NXevent_data)
    events['event_time_offset'] = sc.arange('ignored', 2)
    events['event_time_zero'] = sc.arange('ignored', 2)
    events['event_index'] = sc.arange('ignored', 2)
    events['event_id'] = sc.arange('ignored', 2)
    event_data = nxroot['entry/events_0']
    assert event_data['event_time_offset'].dims == ('event',)
    assert event_data['event_time_zero'].dims == ('event_time_zero',)
    assert event_data['event_index'].dims == ('event_time_zero',)
    assert event_data['event_id'].dims == ('event',)
    log = nxroot['entry'].create_class('log', NXlog)
    log['value'] = sc.arange('ignored', 2)
    log['time'] = sc.arange('ignored', 2)
    assert log['time'].dims == ('time',)
    assert log['value'].dims == ('time',)


def test_field_unit_is_none_if_no_units_attribute(nxroot) -> None:
    log = nxroot.create_class('log', NXlog)
    log['value'] = sc.arange('ignored', 2, unit=None)
    log['time'] = sc.arange('ignored', 2)
    assert log.unit is None
    field = log['value']
    assert field.unit is None


def test_field_errors_with_same_unit_handles_them_with_value(nxroot) -> None:
    entry = nxroot.create_class('group', snx.NXentry)
    entry['value'] = sc.array(dims=['ignored'], values=[10.0], unit='m')
    entry['value_errors'] = sc.array(dims=['ignored'], values=[2.0], unit='m')
    value = nxroot['group']['value'][()]
    assert_identical(value, sc.scalar(value=10.0, variance=4.0, unit='m'))


def test_field_errors_with_different_unit_handles_them_individually(nxroot) -> None:
    entry = nxroot.create_class('group', snx.NXentry)
    entry['value'] = sc.array(dims=['ignored'], values=[10.0], unit='m')
    entry['value_errors'] = sc.array(dims=['ignored'], values=[200.0], unit='cm')
    value = nxroot['group']['value'][()]
    assert_identical(value, sc.scalar(value=10.0, unit='m'))
    assert 'value_errors' in nxroot['group']
    errors = nxroot['group']['value_errors'][()]
    assert_identical(errors, sc.scalar(value=200.0, unit='cm'))


@pytest.mark.parametrize(
    ('value', 'type_'),
    [(1.2, np.float32), (123, np.int32), ('abc', str)],
)
def test_field_is_returned_as_python_object_if_shape_empty_and_no_unit(
    nxroot, value, type_
):
    nxroot['field1'] = sc.scalar(value, unit=None, dtype=type_)
    field = nxroot['field1'][()]
    assert isinstance(field, type_)
    assert field == type_(value)


def test_field_is_returned_as_python_object_if_shape_empty_and_no_unit_bool(nxroot):
    nxroot['field1'] = sc.scalar(True, unit=None, dtype=bool)
    field = nxroot['field1'][()]
    # could be either type depending on Scipp version
    assert isinstance(field, (np.bool_ | bool))
    assert field


@pytest.mark.parametrize('value', [1.2, 123, True, 'abc'])
def test_field_is_returned_as_variable_if_shape_empty_and_unit(nxroot, value) -> None:
    nxroot['field1'] = sc.scalar(value, unit='K')
    field = nxroot['field1'][()]
    assert isinstance(field, sc.Variable)


def test_field_getitem_returns_variable_with_correct_size_and_values(nxroot) -> None:
    nxroot['field'] = sc.arange('ignored', 6, dtype='int64', unit='ns')
    field = nxroot['field']
    assert sc.identical(
        field[...],
        sc.array(dims=['dim_0'], unit='ns', values=[0, 1, 2, 3, 4, 5], dtype='int64'),
    )
    assert sc.identical(
        field[1:],
        sc.array(dims=['dim_0'], unit='ns', values=[1, 2, 3, 4, 5], dtype='int64'),
    )
    assert sc.identical(
        field[:-1],
        sc.array(dims=['dim_0'], unit='ns', values=[0, 1, 2, 3, 4], dtype='int64'),
    )


@pytest.mark.parametrize("string", UTF8_TEST_STRINGS)
def test_field_of_utf8_encoded_dataset_is_loaded_correctly(nxroot, string) -> None:
    nxroot['entry']['title'] = sc.array(
        dims=['ignored'], values=[string, string + string]
    )
    title = nxroot['entry/title']
    assert sc.identical(
        title[...], sc.array(dims=['dim_0'], values=[string, string + string])
    )


@pytest.mark.filterwarnings("ignore:Encoding for bytes")
def test_field_of_extended_ascii_in_ascii_encoded_dataset_is_loaded_correctly() -> None:
    # When writing, if we use bytes h5py will write as ascii encoding
    # 0xb0 = degrees symbol in latin-1 encoding.
    string = b"run at rot=90" + bytes([0xB0])
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        f['title'] = np.array([string, string + b'x'])
        title = snx.Group(f)['title']
        assert sc.identical(
            title[...],
            sc.array(dims=['dim_0'], values=["run at rot=90°", "run at rot=90°x"]),
        )


def test_ms_field_with_second_datetime_attribute_loaded_as_ms_datetime(nxroot) -> None:
    nxroot['mytime'] = sc.arange('ignored', 2, unit='ms')
    nxroot['mytime'].dataset.attrs['start_time'] = '2022-12-12T12:13:14'
    assert sc.identical(
        nxroot['mytime'][...],
        sc.datetimes(
            dims=['dim_0'],
            unit='ms',
            values=['2022-12-12T12:13:14.000', '2022-12-12T12:13:14.001'],
        ),
    )


def test_ns_field_with_second_datetime_attribute_loaded_as_ns_datetime(nxroot) -> None:
    nxroot['mytime'] = sc.arange('ignored', 2, unit='ns')
    nxroot['mytime'].dataset.attrs['start_time'] = '1970-01-01T00:00:00'
    assert sc.identical(
        nxroot['mytime'][...],
        sc.datetimes(
            dims=['dim_0'],
            unit='ns',
            values=['1970-01-01T00:00:00.000000000', '1970-01-01T00:00:00.000000001'],
        ),
    )


def test_second_field_with_ns_datetime_attribute_loaded_as_ns_datetime(nxroot) -> None:
    nxroot['mytime'] = sc.arange('ignored', 2, unit='s')
    nxroot['mytime'].dataset.attrs['start_time'] = '1984-01-01T00:00:00.000000000'
    assert sc.identical(
        nxroot['mytime'][...],
        sc.datetimes(
            dims=['dim_0'],
            unit='ns',
            values=['1984-01-01T00:00:00', '1984-01-01T00:00:01'],
        ),
    )


@pytest.mark.parametrize(
    ('timezone', 'hhmm'),
    [
        ('Z', '12:00'),
        ('+04', '08:00'),
        ('+00', '12:00'),
        ('-02', '14:00'),
        ('+1130', '00:30'),
        ('-0930', '21:30'),
        ('+11:30', '00:30'),
        ('-09:30', '21:30'),
    ],
)
def test_timezone_information_in_datetime_attribute_is_applied(
    nxroot, timezone, hhmm
) -> None:
    nxroot['mytime'] = sc.scalar(value=3, unit='s')
    nxroot['mytime'].dataset.attrs['start_time'] = f'1984-01-01T12:00:00{timezone}'
    assert sc.identical(
        nxroot['mytime'][...], sc.datetime(unit='s', value=f'1984-01-01T{hhmm}:03')
    )


def test_timezone_information_in_datetime_attribute_preserves_ns_precision(
    nxroot,
) -> None:
    nxroot['mytime'] = sc.scalar(value=3, unit='s')
    nxroot['mytime'].dataset.attrs['start_time'] = '1984-01-01T12:00:00.123456789+0200'
    assert sc.identical(
        nxroot['mytime'][...],
        sc.datetime(unit='ns', value='1984-01-01T10:00:03.123456789'),
    )


def test_loads_bare_timestamps_if_multiple_candidate_datetime_offsets_found(
    nxroot,
) -> None:
    offsets = sc.arange('ignored', 2, unit='ms')
    nxroot['mytime'] = offsets
    nxroot['mytime'].dataset.attrs['offset'] = '2022-12-12T12:13:14'
    nxroot['mytime'].dataset.attrs['start_time'] = '2022-12-12T12:13:15'
    assert sc.identical(nxroot['mytime'][...], offsets.rename(ignored='dim_0'))


def test_length_0_field_with_datetime_attribute_loaded_as_datetime(nxroot) -> None:
    nxroot['mytime'] = sc.arange('ignored', 0, unit='ms')
    nxroot['mytime'].dataset.attrs['start_time'] = '2022-12-12T12:13:14'
    assert_identical(
        nxroot['mytime'][...], sc.datetimes(dims=['dim_0'], unit='ms', values=[])
    )


@pytest.mark.skip(reason='Special attributes disabled for now. Do we keep them?')
def test___getattr__for_unique_child_groups(nxroot) -> None:
    entry = nxroot['entry']
    with pytest.raises(NexusStructureError):
        entry.log
    entry.create_class('log1', NXlog)
    log = entry.log
    assert log.nx_class == NXlog
    assert log.name == '/entry/log1'
    assert isinstance(log, NXlog)
    entry.create_class('log2', NXlog)
    with pytest.raises(NexusStructureError):
        entry.log


@pytest.mark.skip(reason='Special attributes disabled for now. Do we keep them?')
def test___dir__(nxroot) -> None:
    entry = nxroot['entry']
    assert 'log' not in entry.__dir__()
    entry.create_class('log1', NXlog)
    assert 'log' in entry.__dir__()
    entry.create_class('log2', NXlog)
    assert 'log' not in entry.__dir__()


def test___dir__includes_non_dynamic_properties(nxroot) -> None:
    entry = nxroot['entry']
    det = entry.create_class('det', NXdetector)
    det.create_class('events', NXevent_data)
    # Ensure we are not replacing __dir__ but adding to it
    assert 'unit' in det.__dir__()


def test_read_recursive(h5root) -> None:
    entry = h5root.create_group('entry')
    data = entry.create_group('data')
    data['signal'] = np.arange(4)
    data['signal'].attrs['units'] = 'm'
    data['time'] = np.arange(5)
    data['time'].attrs['units'] = 's'
    obj = snx.Group(entry)
    dg = obj[()]
    assert obj.sizes == {'dim_0': None}
    assert 'data' in dg


def test_errors_read_as_variances(h5root) -> None:
    entry = h5root.create_group('entry')
    data = entry.create_group('data')
    data['signal'] = np.arange(4.0)
    data['signal'].attrs['units'] = 'm'
    data['signal_errors'] = np.arange(4.0)
    data['signal_errors'].attrs['units'] = 'm'
    data['time'] = np.arange(5.0)
    data['time'].attrs['units'] = 's'
    data['time_errors'] = np.arange(5.0)
    data['time_errors'].attrs['units'] = 's'
    obj = snx.Group(data)
    assert set(obj._children.keys()) == {'signal', 'time'}
    dg = obj[()]
    assert dg['signal'].variances is not None
    assert dg['time'].variances is not None
    assert np.array_equal(dg['signal'].variances, np.arange(4.0) ** 2)
    assert np.array_equal(dg['time'].variances, np.arange(5.0) ** 2)


def test_does_not_require_unit_of_errors(h5root) -> None:
    entry = h5root.create_group('entry')
    data = entry.create_group('data')
    data['signal'] = np.arange(4.0)
    data['signal'].attrs['units'] = 'm'
    data['signal_errors'] = np.arange(4.0)
    # no units on signal_errors
    data['time'] = np.arange(5.0)
    data['time'].attrs['units'] = 's'
    data['time_errors'] = np.arange(5.0)
    # no units on time_errors
    obj = snx.Group(data)
    assert set(obj._children.keys()) == {'signal', 'time'}
    dg = obj[()]
    assert dg['signal'].unit == 'm'
    assert dg['time'].unit == 's'


def test_read_field(h5root) -> None:
    entry = h5root.create_group('entry')
    data = entry.create_group('data')
    data['signal'] = np.arange(4)
    data['signal'].attrs['units'] = 'm'
    obj = snx.Group(data)
    var = obj['signal'][()]
    assert sc.identical(var, sc.array(dims=['dim_0'], values=np.arange(4), unit='m'))


def test_nxdata_with_signal_axes_indices_reads_as_data_array(h5root) -> None:
    entry = h5root.create_group('entry')
    data = entry.create_group('data')
    data.attrs['NX_class'] = 'NXdata'
    data.attrs['signal'] = 'signal'
    data.attrs['axes'] = ['time', 'temperature']
    data.attrs['time_indices'] = [0]
    data.attrs['temperature_indices'] = [1]
    ref = sc.DataArray(
        data=sc.ones(dims=['time', 'temperature'], shape=[3, 4], unit='m')
    )
    ref.coords['time'] = sc.array(dims=['time'], values=np.arange(3), unit='s')
    ref.coords['temperature'] = sc.array(
        dims=['temperature'], values=np.arange(4), unit='K'
    )
    data['signal'] = ref.values
    data['signal'].attrs['units'] = str(ref.unit)
    data['time'] = ref.coords['time'].values
    data['time'].attrs['units'] = str(ref.coords['time'].unit)
    data['temperature'] = ref.coords['temperature'].values
    data['temperature'].attrs['units'] = str(ref.coords['temperature'].unit)
    obj = snx.Group(data, definitions=snx.base_definitions())
    da = obj[()]
    assert sc.identical(da, ref)


def test_nxdata_positional_indexing_returns_correct_slice(h5root) -> None:
    entry = h5root.create_group('entry')
    data = entry.create_group('data')
    data.attrs['NX_class'] = 'NXdata'
    data.attrs['signal'] = 'signal'
    data.attrs['axes'] = ['time', 'temperature']
    data.attrs['time_indices'] = [0]
    data.attrs['temperature_indices'] = [1]
    ref = sc.DataArray(
        data=sc.ones(dims=['time', 'temperature'], shape=[3, 4], unit='m')
    )
    ref.coords['time'] = sc.array(dims=['time'], values=np.arange(3), unit='s')
    ref.coords['temperature'] = sc.array(
        dims=['temperature'], values=np.arange(4), unit='K'
    )
    data['signal'] = ref.values
    data['signal'].attrs['units'] = str(ref.unit)
    data['time'] = ref.coords['time'].values
    data['time'].attrs['units'] = str(ref.coords['time'].unit)
    data['temperature'] = ref.coords['temperature'].values
    data['temperature'].attrs['units'] = str(ref.coords['temperature'].unit)
    obj = snx.Group(data, definitions=snx.base_definitions())
    da = obj['time', 0:2]
    assert sc.identical(da, ref['time', 0:2])


def test_nxdata_label_indexing_returns_correct_slice(h5root) -> None:
    entry = h5root.create_group('entry')
    data = entry.create_group('data')
    data.attrs['NX_class'] = 'NXdata'
    data.attrs['signal'] = 'signal'
    data.attrs['axes'] = ['time', 'temperature']
    data.attrs['time_indices'] = [0]
    data.attrs['temperature_indices'] = [1]
    ref = sc.DataArray(
        data=sc.ones(dims=['time', 'temperature'], shape=[3, 4], unit='m')
    )
    ref.coords['time'] = sc.array(dims=['time'], values=np.arange(3), unit='s')
    ref.coords['temperature'] = sc.array(
        dims=['temperature'], values=np.arange(4), unit='K'
    )
    data['signal'] = ref.values
    data['signal'].attrs['units'] = str(ref.unit)
    data['time'] = ref.coords['time'].values
    data['time'].attrs['units'] = str(ref.coords['time'].unit)
    data['temperature'] = ref.coords['temperature'].values
    data['temperature'].attrs['units'] = str(ref.coords['temperature'].unit)
    obj = snx.Group(data, definitions=snx.base_definitions())
    da = obj['time', sc.scalar(0, unit='s') : sc.scalar(2, unit='s')]
    assert sc.identical(
        da, ref['time', sc.scalar(0, unit='s') : sc.scalar(2, unit='s')]
    )


def test_nxdata_with_bin_edges_positional_indexing_returns_correct_slice(
    h5root,
) -> None:
    entry = h5root.create_group('entry')
    data = entry.create_group('data')
    data.attrs['NX_class'] = 'NXdata'
    data.attrs['signal'] = 'signal'
    data.attrs['axes'] = ['time', 'temperature']
    data.attrs['time_indices'] = [0]
    data.attrs['temperature_indices'] = [1]
    ref = sc.DataArray(
        data=sc.ones(dims=['time', 'temperature'], shape=[3, 4], unit='m')
    )
    ref.coords['time'] = sc.array(dims=['time'], values=np.arange(3), unit='s')
    ref.coords['temperature'] = sc.array(
        dims=['temperature'], values=np.arange(5), unit='K'
    )
    data['signal'] = ref.values
    data['signal'].attrs['units'] = str(ref.unit)
    data['time'] = ref.coords['time'].values
    data['time'].attrs['units'] = str(ref.coords['time'].unit)
    data['temperature'] = ref.coords['temperature'].values
    data['temperature'].attrs['units'] = str(ref.coords['temperature'].unit)
    obj = snx.Group(data, definitions=snx.base_definitions())
    da = obj['temperature', 0:2]
    assert sc.identical(da, ref['temperature', 0:2])


def test_nxdata_with_bin_edges_label_indexing_returns_correct_slice(h5root) -> None:
    entry = h5root.create_group('entry')
    data = entry.create_group('data')
    data.attrs['NX_class'] = 'NXdata'
    data.attrs['signal'] = 'signal'
    data.attrs['axes'] = ['time', 'temperature']
    data.attrs['time_indices'] = [0]
    data.attrs['temperature_indices'] = [1]
    ref = sc.DataArray(
        data=sc.ones(dims=['time', 'temperature'], shape=[3, 4], unit='m')
    )
    ref.coords['time'] = sc.array(dims=['time'], values=np.arange(3), unit='s')
    ref.coords['temperature'] = sc.array(
        dims=['temperature'], values=np.arange(5), unit='K'
    )
    data['signal'] = ref.values
    data['signal'].attrs['units'] = str(ref.unit)
    data['time'] = ref.coords['time'].values
    data['time'].attrs['units'] = str(ref.coords['time'].unit)
    data['temperature'] = ref.coords['temperature'].values
    data['temperature'].attrs['units'] = str(ref.coords['temperature'].unit)
    obj = snx.Group(data, definitions=snx.base_definitions())
    da = obj['temperature', sc.scalar(0, unit='K') : sc.scalar(2, unit='K')]
    assert sc.identical(
        da, ref['temperature', sc.scalar(0, unit='K') : sc.scalar(2, unit='K')]
    )


def create_nexus_group_with_data_arrays(h5root, dims, coords):
    entry = h5root.create_group('entry')
    for i, (dim, coord) in enumerate(zip(dims, coords, strict=True)):
        data = entry.create_group(f'data_{i}')
        data.attrs['NX_class'] = 'NXdata'
        data['signal'] = np.arange(5)
        data['signal'].attrs['units'] = 'm'
        data[coord] = np.arange(5)
        data[coord].attrs['units'] = 's'
        data.attrs['signal'] = 'signal'
        data.attrs['axes'] = [dim]
        data.attrs[f'{dim}_indices'] = [0]
    return snx.Group(h5root, definitions=snx.base_definitions())


@pytest.mark.parametrize(
    "slice_",
    [
        ('time', sc.scalar(1, unit='s')),
        ('time', slice(sc.scalar(1, unit='s'), None)),
        ('time', slice(None, sc.scalar(1, unit='s'))),
        {'time': sc.scalar(1, unit='s'), 'x': sc.scalar(1, unit='s')},
    ],
)
@pytest.mark.parametrize(
    ('dims', 'coords'),
    [
        (('time',), ('time2',)),
        (('time', 'time'), ('time2', 'time')),
    ],
)
def test_label_indexing_group_behaves_same_as_indexing_scipp_datagroup(
    h5root, slice_, dims, coords
):
    nx = create_nexus_group_with_data_arrays(h5root, dims, coords)
    dg = nx[()]

    exception = None
    try:
        # Scipp does not support dict slicing,
        # manually slice datagroup in multiple coords
        if isinstance(slice_, dict):
            for s in slice_.items():
                dg = dg[s]
        else:
            dg = dg[slice_]
    except Exception as e:
        exception = type(e)

    if exception:
        with pytest.raises(exception):
            nx[slice_]
    else:
        assert_identical(nx[slice_], dg)


def test_create_field_saves_errors(nxroot) -> None:
    entry = nxroot['entry']
    data = sc.array(
        dims=['d0'], values=[1.2, 3.4, 5.6], variances=[0.9, 0.8, 0.7], unit='cm'
    )
    entry.create_field('signal', data)

    loaded = entry['signal'][()]
    # Use allclose instead of identical because the variances are stored as stddevs
    # which loses precision.
    assert sc.allclose(loaded, data.rename_dims(d0='dim_0'))


@pytest.mark.parametrize('nxclass', [NXlog, NXmonitor, NXdetector, NXevent_data])
def test_empty_class_does_not_warn(nxroot, nxclass) -> None:
    log = nxroot['entry'].create_class('log', nxclass)
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        log[()]


def test_trailing_forward_slash_in_path_does_not_change_file_object(nxroot) -> None:
    assert id(nxroot['entry/']) == id(nxroot['entry'])


def test_path_santization(nxroot) -> None:
    nxroot['entry'].create_class('log', NXlog)
    assert id(nxroot['/entry/log']) == id(nxroot['/entry//log'])
    assert id(nxroot['entry/log']) == id(nxroot['entry//log'])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
# ruff: noqa: E402, F401, I

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

from . import typing
from ._load import load
from .attrs import Attrs
from .base import (
    Group,
    NexusStructureError,
    NXobject,
    base_definitions,
    create_class,
    create_field,
)
from .field import DependsOn, Field
from .file import File
from .nexus_classes import *  # noqa: F403
from .nxtransformations import TransformationChain, compute_positions, zip_pixel_offsets

__all__ = [
    'Attrs',
    'DependsOn',
    'Field',
    'File',
    'Group',
    'NXobject',
    'NexusStructureError',
    'TransformationChain',
    'base_definitions',
    'compute_positions',
    'create_class',
    'create_field',
    'load',
    'typing',
    'zip_pixel_offsets',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import h5py
import numpy as np
import pytest
import scipp as sc

import scippnexus as snx


@pytest.fixture
def h5root():
    """Yield h5py root group (file)"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        yield f


@pytest.fixture
def nxroot():
    """Yield NXroot containing a single NXentry named 'entry'"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        root = snx.Group(f, definitions=snx.base_definitions())
        root.create_class('entry', snx.NXentry)
        yield root


def create_event_data_ids_1234(group):
    group['event_id'] = sc.array(dims=[''], unit=None, values=[1, 2, 4, 1, 2, 2])
    group['event_time_offset'] = sc.array(
        dims=[''], unit='s', values=[456, 7, 3, 345, 632, 23]
    )
    group['event_time_zero'] = sc.array(dims=[''], unit='s', values=[1, 2, 3, 4])
    group['event_index'] = sc.array(dims=[''], unit=None, values=[0, 3, 3, -1000])


def test_negative_event_index_converted_to_num_event(nxroot) -> None:
    event_data = nxroot['entry'].create_class('events_0', snx.NXevent_data)
    create_event_data_ids_1234(event_data)
    events = nxroot['entry/events_0'][...]
    assert events.bins.size().values[2] == 3
    assert events.bins.size().values[3] == 0


@pytest.mark.filterwarnings("ignore:Failed to load /entry/events_0:UserWarning")
def test_bad_event_index_causes_load_as_DataGroup(nxroot) -> None:
    event_data = nxroot['entry'].create_class('events_0', snx.NXevent_data)
    event_data['event_id'] = sc.array(dims=[''], unit=None, values=[1, 2, 4, 1, 2])
    event_data['event_time_offset'] = sc.array(dims=[''], unit='s', values=[0, 0, 0, 0])
    event_data['event_time_zero'] = sc.array(dims=[''], unit='s', values=[1, 2, 3, 4])
    event_data['event_index'] = sc.array(dims=[''], unit=None, values=[0, 3, 3, 666])
    dg = nxroot['entry/events_0'][...]
    assert isinstance(dg, sc.DataGroup)


def create_event_data_without_event_id(group):
    group['event_time_offset'] = sc.array(
        dims=[''], unit='s', values=[456, 7, 3, 345, 632, 23]
    )
    group['event_time_zero'] = sc.array(dims=[''], unit='s', values=[1, 2, 3, 4])
    group['event_index'] = sc.array(dims=[''], unit=None, values=[0, 3, 3, 5])


def test_event_data_without_event_id_can_be_loaded(nxroot) -> None:
    event_data = nxroot['entry'].create_class('events_0', snx.NXevent_data)
    create_event_data_without_event_id(event_data)
    da = event_data[...]
    assert len(da.bins.coords) == 1
    assert 'event_time_offset' in da.bins.coords


def create_event_data_without_event_time_zero(group):
    group['event_id'] = sc.array(dims=[''], unit=None, values=[1, 2, 4, 1, 2, 2])
    group['event_time_offset'] = sc.array(
        dims=[''], unit='s', values=[456, 7, 3, 345, 632, 23]
    )
    group['event_index'] = sc.array(dims=[''], unit=None, values=[0, 3, 3, 5])


def test_event_data_without_event_time_zero_can_be_loaded(nxroot) -> None:
    event_data = nxroot['entry'].create_class('events_0', snx.NXevent_data)
    create_event_data_without_event_time_zero(event_data)
    da = event_data[...]
    assert len(da.bins.coords) == 2
    assert 'event_time_offset' in da.bins.coords


def test_event_mode_monitor_without_event_id_can_be_loaded(nxroot) -> None:
    monitor = nxroot['entry'].create_class('monitor', snx.NXmonitor)
    create_event_data_without_event_id(monitor)
    da = monitor[...]['events']
    assert 'event_time_offset' in da.bins.coords


def test_read_empty_nxevent_data(h5root) -> None:
    entry = h5root.create_group('entry')
    events = entry.create_group('events')
    events.attrs['NX_class'] = 'NXevent_data'
    root = snx.Group(entry)
    event_data = root['events']
    dg = event_data[()]
    assert sc.identical(dg, sc.DataGroup())


def make_event_data(h5root):
    entry = h5root.create_group('entry')
    events = entry.create_group('events')
    events.attrs['NX_class'] = 'NXevent_data'
    rng = np.random.default_rng(0)
    events['event_id'] = rng.integers(0, 2, size=4)
    events['event_time_offset'] = np.arange(4)
    events['event_time_offset'].attrs['units'] = 'ns'
    events['event_time_zero'] = np.array([100, 200])
    events['event_time_zero'].attrs['units'] = 'ms'
    events['event_index'] = np.array([0, 3])
    return entry


def test_nxevent_data_keys(h5root) -> None:
    entry = make_event_data(h5root)
    root = snx.Group(entry)
    event_data = root['events']
    assert set(event_data.keys()) == {
        'event_id',
        'event_time_offset',
        'event_time_zero',
        'event_index',
    }


def test_nxevent_data_children_read_as_variables_with_correct_dims(h5root) -> None:
    entry = make_event_data(h5root)
    root = snx.Group(entry, definitions=snx.base_definitions())
    event_data = root['events']
    assert sc.identical(
        event_data['event_id'][()],
        sc.array(dims=['event'], values=[1, 1, 1, 0], unit=None),
    )
    assert sc.identical(
        event_data['event_time_offset'][()],
        sc.array(dims=['event'], values=[0, 1, 2, 3], unit='ns'),
    )
    assert sc.identical(
        event_data['event_time_zero'][()],
        sc.array(dims=['event_time_zero'], values=[100, 200], unit='ms'),
    )
    assert sc.identical(
        event_data['event_index'][()],
        sc.array(dims=['event_time_zero'], values=[0, 3], unit=None),
    )


def test_nxevent_data_dims_and_sizes_ignore_pulse_contents(h5root) -> None:
    entry = make_event_data(h5root)
    root = snx.Group(entry, definitions=snx.base_definitions())
    event_data = root['events']
    assert event_data.dims == ('event_time_zero',)
    assert event_data.sizes == {'event_time_zero': 2}


def test_read_nxevent_data(h5root) -> None:
    entry = make_event_data(h5root)
    root = snx.Group(entry, definitions=snx.base_definitions())
    event_data = root['events']
    da = event_data[()]
    assert sc.identical(
        da.data.bins.size(),
        sc.array(dims=['event_time_zero'], values=[3, 1], unit=None),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Matthew Jones
import warnings

import h5py


def cset_to_encoding(cset: int) -> str:
    """
    Converts a HDF5 cset into a python encoding. Allowed values for cset are
    h5py.h5t.CSET_ASCII and h5py.h5t.CSET_UTF8.

    Args:
        cset: The HDF character set to convert

    Returns:
        A string describing the encoding suitable for calls to str(encoding=...)
        Either "ascii" or "utf-8".
    """
    if cset == h5py.h5t.CSET_ASCII:
        return "ascii"
    elif cset == h5py.h5t.CSET_UTF8:
        return "utf-8"
    else:
        raise ValueError(
            f"Unknown character set in HDF5 data file. Expected data "
            f"types are {h5py.h5t.CSET_ASCII=} or "
            f"{h5py.h5t.CSET_UTF8=} but got '{cset}'. "
        )


def warn_latin1_decode(obj: object, decoded: str, error: Exception) -> None:
    warnings.warn(
        f"Encoding for bytes '{obj!r}' declared as ascii, "
        f"but contains characters in extended ascii range. Assuming "
        f"extended ASCII (latin-1), but this behavior is not "
        f"specified by the HDF5 or nexus standards and may therefore "
        f"be incorrect. Decoded string using latin-1 is '{decoded}'. "
        f"Error was '{error}'.",
        stacklevel=2,
    )


def ensure_str(str_or_bytes: str | bytes, encoding: str) -> str:
    """
    See https://docs.h5py.org/en/stable/strings.html for justification about some of
    the operations performed in this method. In particular, variable-length strings
    are returned as `str` from h5py, but need to be encoded using the surrogateescape
    error handler and then decoded using the encoding specified in the nexus file in
    order to get a correctly encoded string in all cases.

    Note that the nexus standard leaves unspecified the behavior of H5T_CSET_ASCII
    for characters >=128. Common extensions are the latin-1 ("extended ascii") character
    set which appear to be used in nexus files from some facilities. Attempt to load
    these strings with the latin-1 extended character set, but warn as this is
    technically unspecified behavior.
    """
    if isinstance(str_or_bytes, str):
        str_or_bytes = str_or_bytes.encode("utf-8", errors="surrogateescape")

    if encoding == "ascii":
        try:
            return str(str_or_bytes, encoding="ascii")
        except UnicodeDecodeError as e:
            decoded = str(str_or_bytes, encoding="latin-1")
            warn_latin1_decode(str_or_bytes, decoded, e)
            return decoded
    else:
        return str(str_or_bytes, encoding)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from collections.abc import Iterator, Mapping
from typing import Any

import h5py as h5

from ._string import cset_to_encoding, ensure_str


class Attrs(Mapping[str, Any]):
    def __init__(self, attrs: h5.AttributeManager) -> None:
        self._base_attrs = attrs
        self._attrs = dict(attrs)

    def __getitem__(self, name: str) -> Any:
        attr = self._attrs[name]
        # Is this check for string attributes sufficient? Is there a better way?
        if isinstance(attr, str | bytes):
            cset = self._base_attrs.get_id(name.encode("utf-8")).get_type().get_cset()
            return ensure_str(attr, cset_to_encoding(cset))
        return attr

    def __iter__(self) -> Iterator[str]:
        return iter(self._attrs)

    def __len__(self) -> int:
        return len(self._attrs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import os
import sys
from pathlib import Path
from typing import Any

import pytest
import scipp as sc

# Silence warning from Jupyter
os.environ['JUPYTER_PLATFORM_DIRS'] = '1'
# To import helper modules
sys.path.insert(0, str(Path(__file__).resolve().parent))

pytest.register_assert_rewrite('scipp.testing.assertions')


def pytest_assertrepr_compare(op: str, left: Any, right: Any) -> list[str]:
    if isinstance(left, sc.Unit) and isinstance(right, sc.Unit):
        return [f'Unit({left}) {op} Unit({right})']
    if isinstance(left, sc.DType) or isinstance(right, sc.DType):
        return [f'{left!r} {op} {right!r}']



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import io
from pathlib import Path

import h5py as h5
import pytest
import scipp as sc
import scipp.testing

import scippnexus as snx


@pytest.fixture
def reference_data() -> sc.DataGroup:
    return sc.DataGroup(
        entry=sc.DataGroup(
            log=sc.DataArray(
                data=sc.arange('time', -0.5, 1.5, 0.5, unit='rad'),
                coords={
                    'time': sc.array(
                        dims=['time'], values=[100, 200, 300, 400], unit='s'
                    )
                    + sc.epoch(unit='s')
                },
            ),
            instrument=sc.DataGroup(
                detector=sc.DataGroup(
                    data=sc.DataArray(
                        data=sc.array(
                            dims=['detector_number'], values=[24, 56, 19], unit='counts'
                        ),
                        coords={
                            'detector_number': sc.array(
                                dims=['detector_number'], values=[1, 3, 4], unit=None
                            )
                        },
                    )
                )
            ),
        )
    )


@pytest.fixture
def nexus_buffer(reference_data) -> io.BytesIO:
    buffer = io.BytesIO()
    with snx.File(buffer, 'w') as root:
        entry = root.create_class('entry', snx.NXentry)

        ref_log = reference_data['entry']['log']
        log = entry.create_class('log', snx.NXlog)
        log.create_field('value', ref_log.data)
        log.create_field('time', ref_log.coords['time'])

        ref_data = reference_data['entry']['instrument']['detector']['data']
        instrument = entry.create_class('instrument', snx.NXinstrument)
        detector = instrument.create_class('detector', snx.NXdetector)
        detector.create_field('detector_number', ref_data.coords['detector_number'])
        detector.create_field('data', ref_data.data)
    return buffer


def test_load_all_from_buffer(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    loaded = snx.load(nexus_buffer)
    sc.testing.assert_identical(loaded, reference_data)


@pytest.mark.parametrize('path_type', [str, Path])
def test_load_all_from_file(
    path_type: type,
    nexus_buffer: io.BytesIO,
    reference_data: sc.DataGroup,
    tmp_path: Path,
) -> None:
    filename = tmp_path / 'test.nxs'
    filename.write_bytes(nexus_buffer.getvalue())
    loaded = snx.load(path_type(filename))
    sc.testing.assert_identical(loaded, reference_data)


def test_load_all_from_snx_group(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    with snx.File(nexus_buffer, 'r') as f:
        loaded = snx.load(f)
    sc.testing.assert_identical(loaded, reference_data)


def test_load_all_from_nested_snx_group(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    with snx.File(nexus_buffer, 'r') as f:
        loaded = snx.load(f['entry'])
    sc.testing.assert_identical(loaded, reference_data['entry'])


def test_load_all_from_h5_group(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    with h5.File(nexus_buffer, 'r') as f:
        loaded = snx.load(f)
    sc.testing.assert_identical(loaded, reference_data)


def test_load_all_from_nested_h5_group(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    with h5.File(nexus_buffer, 'r') as f:
        loaded = snx.load(f['entry'])
    sc.testing.assert_identical(loaded, reference_data['entry'])


def test_load_from_root_single_segment(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    loaded = snx.load(nexus_buffer, root='entry')
    sc.testing.assert_identical(loaded, reference_data['entry'])


def test_load_from_root_multiple_segments(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    loaded = snx.load(nexus_buffer, root='entry/instrument/detector')
    sc.testing.assert_identical(
        loaded, reference_data['entry']['instrument']['detector']
    )


def test_load_nonexistent_root(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    with pytest.raises(KeyError):
        snx.load(nexus_buffer, root='entry/does-not-exist')


def test_load_select_index_range(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    loaded = snx.load(nexus_buffer, select={'time': slice(1, None)})
    sc.testing.assert_identical(loaded, reference_data['time', 1:])


def test_load_select_index(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    loaded = snx.load(nexus_buffer, select={'time': 2})
    sc.testing.assert_identical(loaded, reference_data['time', 2])


def test_load_from_root_select_index_range(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    loaded = snx.load(
        nexus_buffer,
        root='entry/instrument',
        select={'detector_number': slice(None, 2)},
    )
    sc.testing.assert_identical(
        loaded, reference_data['entry']['instrument']['detector_number', :2]
    )


def test_load_select_unknown_dim(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    with pytest.raises(sc.DimensionError):
        snx.load(nexus_buffer, select={'y': 3})


class NXdetectorTimes10(snx.NXdetector):
    def assemble(self, dg: sc.DataGroup) -> sc.DataGroup:
        return 10 * super().assemble(dg)


def test_load_from_buffer_with_explicit_base_definitions(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    loaded = snx.load(nexus_buffer, definitions=snx.base_definitions())
    sc.testing.assert_identical(loaded, reference_data)


def test_load_from_buffer_with_custom_definitions(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    definitions = {**snx.base_definitions(), 'NXdetector': NXdetectorTimes10}
    loaded = snx.load(nexus_buffer, definitions=definitions)
    expected = reference_data.copy()
    expected['entry']['instrument']['detector']['data'] *= 10
    sc.testing.assert_identical(loaded, expected)


def test_load_from_h5_group_with_custom_definitions(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    definitions = {**snx.base_definitions(), 'NXdetector': NXdetectorTimes10}
    with h5.File(nexus_buffer, 'r') as f:
        loaded = snx.load(f, definitions=definitions)
    expected = reference_data.copy()
    expected['entry']['instrument']['detector']['data'] *= 10
    sc.testing.assert_identical(loaded, expected)


def test_load_from_snx_group_rejects_new_definitions(
    nexus_buffer: io.BytesIO, reference_data: sc.DataGroup
) -> None:
    definitions = {**snx.base_definitions(), 'NXdetector': NXdetectorTimes10}
    with pytest.raises(TypeError, match='Cannot override application definitions'):
        with snx.File(nexus_buffer, 'r') as f:
            snx.load(f, definitions=definitions)



import h5py
import numpy as np
import pytest
import scipp as sc

import scippnexus as snx
from scippnexus.nxoff_geometry import NXoff_geometry, off_to_shape


@pytest.fixture
def group():
    """Yield NXroot containing a single NXentry named 'entry'"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        yield snx.Group(f, definitions=snx.base_definitions())


def test_vertices_loaded_as_vector3(group) -> None:
    shape = group.create_class('shape', NXoff_geometry)
    values = [[1, 2, 3], [4, 5, 6]]
    shape['vertices'] = sc.array(dims=['ignored', 'comp'], values=values, unit='mm')
    loaded = shape[()]
    assert sc.identical(
        loaded['vertices'], sc.vectors(dims=['vertex'], values=values, unit='mm')
    )


def test_field_properties(group) -> None:
    shape = group.create_class('shape', NXoff_geometry)
    values = [[1, 2, 3], [4, 5, 6]]
    shape['vertices'] = sc.array(dims=['ignored', 'comp'], values=values, unit='m')
    shape['winding_order'] = sc.array(dims=['ignored'], values=[], unit=None)
    shape['faces'] = sc.array(dims=['ignored'], values=[], unit=None)
    loaded = shape[()]
    assert loaded['vertices'].dims == ('vertex',)
    assert loaded['winding_order'].dims == ('winding_order',)
    assert loaded['winding_order'].unit is None
    assert loaded['faces'].dims == ('face',)
    assert loaded['faces'].unit is None


def test_off_to_shape_without_detector_faces_yields_scalar_shape_with_all_faces(
    group,
) -> None:
    off = group.create_class('off', NXoff_geometry)
    values = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    off['vertices'] = sc.array(dims=['_', 'comp'], values=values, unit='m')
    off['winding_order'] = sc.array(dims=['_'], values=[0, 1, 2, 0, 2, 1], unit=None)
    off['faces'] = sc.array(dims=['_'], values=[0, 3], unit=None)
    loaded = off[()]
    shape = off_to_shape(**loaded)
    assert shape.ndim == 0
    assert sc.identical(shape.bins.size(), sc.index(2))


def test_off_to_shape_raises_if_detector_faces_but_not_detector_numbers_given(
    group,
) -> None:
    off = group.create_class('off', NXoff_geometry)
    values = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    off['vertices'] = sc.array(dims=['_', 'comp'], values=values, unit='m')
    off['winding_order'] = sc.array(dims=['_'], values=[0, 1, 2, 0, 2, 1], unit=None)
    off['faces'] = sc.array(dims=['_'], values=[0, 3], unit=None)
    det_num1 = 1
    det_num2 = 3
    off['detector_faces'] = sc.array(
        dims=['_', 'dummy'], values=[[0, det_num2], [1, det_num1]], unit=None
    )
    loaded = off[()]
    with pytest.raises(snx.NexusStructureError):
        off_to_shape(**loaded)


def test_off_to_shape_with_single_detector_yields_1d_shape(group) -> None:
    off = group.create_class('off', NXoff_geometry)
    values = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    off['vertices'] = sc.array(dims=['_', 'comp'], values=values, unit='m')
    off['winding_order'] = sc.array(dims=['_'], values=[0, 1, 2, 0, 2, 1], unit=None)
    off['faces'] = sc.array(dims=['_'], values=[0, 3], unit=None)
    det_num1 = 7
    off['detector_faces'] = sc.array(
        dims=['_', 'dummy'], values=[[0, det_num1], [1, det_num1]], unit=None
    )
    loaded = off[()]

    detector_number = sc.index(1)  # not in detector_faces => no faces
    shape = off_to_shape(**loaded, detector_number=detector_number)
    assert sc.identical(shape.bins.size(), sc.array(dims=[], values=0, unit=None))

    detector_number = sc.index(det_num1)
    shape = off_to_shape(**loaded, detector_number=detector_number)
    assert sc.identical(shape.bins.size(), sc.array(dims=[], values=2, unit=None))

    detector_number = sc.array(dims=['detector_number'], values=[det_num1], unit=None)
    shape = off_to_shape(**loaded, detector_number=detector_number)
    assert sc.identical(
        shape.bins.size(), sc.array(dims=['detector_number'], values=[2], unit=None)
    )


def test_off_to_shape_with_two_detectors_yields_1d_shape(group) -> None:
    off = group.create_class('off', NXoff_geometry)
    values = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    off['vertices'] = sc.array(dims=['_', 'comp'], values=values, unit='m')
    off['winding_order'] = sc.array(dims=['_'], values=[0, 1, 2, 0, 2, 1], unit=None)
    off['faces'] = sc.array(dims=['_'], values=[0, 3], unit=None)
    det_num1 = 1
    det_num2 = 3
    off['detector_faces'] = sc.array(
        dims=['_', 'dummy'], values=[[0, det_num2], [1, det_num1]], unit=None
    )
    loaded = off[()]
    detector_number = sc.array(
        dims=['detector_number'], values=[det_num1, det_num2], unit=None
    )
    shape = off_to_shape(**loaded, detector_number=detector_number)
    assert shape.sizes == {'detector_number': 2}
    assert sc.identical(
        shape.bins.size(), sc.array(dims=['detector_number'], values=[1, 1], unit=None)
    )
    assert sc.identical(
        shape[0].value,
        sc.vectors(dims=['face', 'vertex'], values=[values[[0, 2, 1]]], unit='m'),
    )
    assert sc.identical(
        shape[1].value, sc.vectors(dims=['face', 'vertex'], values=[values], unit='m')
    )


def test_off_to_shape_uses_order_of_provided_detector_number_param(group) -> None:
    off = group.create_class('off', NXoff_geometry)
    values = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    off['vertices'] = sc.array(dims=['_', 'comp'], values=values, unit='m')
    off['winding_order'] = sc.array(dims=['_'], values=[0, 1, 2, 0, 2, 1], unit=None)
    off['faces'] = sc.array(dims=['_'], values=[0, 3], unit=None)
    det_num1 = 1
    det_num2 = 3
    off['detector_faces'] = sc.array(
        dims=['_', 'dummy'], values=[[0, det_num2], [1, det_num1]], unit=None
    )
    loaded = off[()]
    detector_number = sc.array(dims=['detector_number'], values=[3, 1], unit=None)
    shape = off_to_shape(**loaded, detector_number=detector_number)
    assert sc.identical(
        shape[0].value, sc.vectors(dims=['face', 'vertex'], values=[values], unit='m')
    )
    assert sc.identical(
        shape[1].value,
        sc.vectors(dims=['face', 'vertex'], values=[values[[0, 2, 1]]], unit='m'),
    )



import h5py
import pytest
import scipp as sc
from scipp.testing import assert_identical

import scippnexus as snx
from scippnexus.application_definitions import nxcansas


@pytest.fixture
def nxroot():
    """Yield NXroot containing a single NXentry named 'entry'"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        root = snx.Group(f, definitions=snx.base_definitions())
        root.create_class('entry', snx.NXentry)
        yield root


def test_setitem_SASentry(nxroot) -> None:
    nxroot['sasentry'] = nxcansas.SASentry(title='A test', run=12345)
    assert 'sasentry' in nxroot
    entry = nxroot['sasentry']
    assert entry.attrs['definition'] == 'NXcanSAS'
    assert entry['title'][()] == 'A test'
    assert entry['run'][()] == 12345


@pytest.fixture
def I_of_Q():
    data = sc.array(
        dims=['Q'],
        values=[0.1, 0.2, 0.1, 0.4],
        variances=[1.0, 4.0, 9.0, 4.0],  # values chosen for exact sqrt
        unit='1/counts',
    )
    da = sc.DataArray(data=data)
    da.coords['Q'] = sc.linspace('Q', 0, 1, num=5, unit='1/angstrom')
    da.coords['Q'].variances = sc.array(dims=['Q'], values=[1, 1, 4, 4, 1]).values
    return da


def test_setitem_SASdata_raises_ValueError_when_given_bin_edges(nxroot, I_of_Q) -> None:
    with pytest.raises(ValueError, match='Q is given as bin-edges'):
        nxroot['sasdata'] = nxcansas.SASdata(I_of_Q, Q_variances='resolutions')


def test_setitem_SASdata(nxroot, I_of_Q) -> None:
    I_of_Q.coords['Q'] = I_of_Q.coords['Q'][1:]
    nxroot['sasdata'] = nxcansas.SASdata(I_of_Q, Q_variances='resolutions')
    data = nxroot['sasdata']
    assert sc.identical(data[...], I_of_Q)


def test_setitem_SASdata_raises_if_interpretation_of_variances_not_specified(
    nxroot,
) -> None:
    data = sc.array(
        dims=['Q'],
        values=[0.1, 0.2, 0.1, 0.4],
        variances=[1.0, 4.0, 9.0, 4.0],  # values chosen for exact sqrt
        unit='1/counts',
    )
    da = sc.DataArray(data=data)
    da.coords['Q'] = sc.linspace('Q', 0, 1, num=4, unit='1/angstrom')
    da.coords['Q'].variances = sc.array(dims=['Q'], values=[1, 4, 4, 1]).values
    with pytest.raises(ValueError, match='Q has variances, must specify whether these'):
        nxroot['sasdata'] = nxcansas.SASdata(da)


def test_load_SASdata(nxroot) -> None:
    nxroot['sasentry'] = nxcansas.SASentry(title='A test', run=12345)
    entry = nxroot['sasentry']
    da = sc.DataArray(
        sc.array(dims=['Q'], values=[1, 2, 3], unit=''),
        coords={'Q': sc.array(dims=['Q'], values=[1, 2, 3, 4], unit='1/angstrom')},
    )
    group = entry.create_class('sasdata', snx.NXdata)
    group._group.attrs['canSAS_class'] = 'SASdata'
    group._group.attrs['signal'] = 'I'
    group._group.attrs['I_axes'] = 'Q'
    group['I'] = da.data
    group['Q'] = da.coords['Q']
    sasroot = snx.Group(nxroot._group, definitions=nxcansas.definitions)
    loaded = sasroot['sasentry/sasdata'][()]
    assert_identical(loaded, da)


def test_load_SAStransmission_spectrum(nxroot) -> None:
    nxroot['sasentry'] = nxcansas.SASentry(title='A test', run=12345)
    entry = nxroot['sasentry']
    spectrum = sc.DataArray(
        sc.array(dims=['lambda'], values=[1, 2, 3], unit='counts'),
        coords={
            'lambda': sc.array(dims=['lambda'], values=[1, 2, 3, 4], unit='angstrom')
        },
    )
    group = entry.create_class('sastransmission_spectrum', snx.NXdata)
    group._group.attrs['canSAS_class'] = 'SAStransmission_spectrum'
    group._group.attrs['signal'] = 'T'
    group._group.attrs['T_axes'] = 'lambda'
    group['T'] = spectrum.data
    group['lambda'] = spectrum.coords['lambda']
    sasroot = snx.Group(nxroot._group, definitions=nxcansas.definitions)
    loaded = sasroot['sasentry/sastransmission_spectrum'][()]
    assert_identical(loaded, spectrum)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import io
from pathlib import Path

import h5py as h5
import pytest

import scippnexus as snx


def is_closed(file: snx.File) -> bool:
    try:
        file.create_class('x', 'NXentry')
    except ValueError as err:
        return 'Unable to synchronously create group' in err.args[0]
    return False


@pytest.mark.parametrize('path_type', [str, Path])
def test_load_entry_from_filename(tmp_path, path_type) -> None:
    with h5.File(tmp_path / 'test.nxs', 'w') as f:
        snx.create_class(f, 'entry', snx.NXentry)

    with snx.File(path_type(tmp_path / 'test.nxs'), 'r') as f:
        assert f.keys() == {'entry'}
    assert is_closed(f)


def test_load_entry_from_buffer(tmp_path) -> None:
    buffer = io.BytesIO()
    with h5.File(buffer, 'w') as f:
        snx.create_class(f, 'entry', snx.NXentry)
    buffer.seek(0)

    with snx.File(buffer, 'r') as f:
        assert f.keys() == {'entry'}
    assert is_closed(f)


def test_load_entry_from_h5py_group_root(tmp_path) -> None:
    with h5.File('test.nxs', 'w', driver='core', backing_store=False) as h5_file:
        snx.create_class(h5_file, 'entry', snx.NXentry)
        with snx.File(h5_file) as snx_file:
            assert snx_file.keys() == {'entry'}
        # Not great, but we don't want to close the h5py file ourselves.
        assert not is_closed(snx_file)


def test_load_entry_from_h5py_group_not_root(tmp_path) -> None:
    with h5.File('test.nxs', 'w', driver='core', backing_store=False) as h5_file:
        entry = snx.create_class(h5_file, 'entry', snx.NXentry)
        snx.create_class(entry, 'user', snx.NXuser)
        with snx.File(h5_file['entry']) as snx_file:
            assert snx_file.keys() == {'user'}
        # Not great, but we don't want to close the h5py file ourselves.
        assert not is_closed(snx_file)


def test_file_from_h5py_group_does_not_allow_extra_args(tmp_path) -> None:
    with h5.File('test.nxs', 'w', driver='core', backing_store=False) as h5_file:
        snx.create_class(h5_file, 'entry', snx.NXentry)
        with pytest.raises(
            TypeError, match='Cannot provide both h5py.File and other arguments'
        ):
            snx.File(h5_file, 'r')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from typing import Any

import scipp as sc

from .base import Group, NXobject, ScippIndex, base_definitions_dict
from .field import Field

_matrix_units = {'orientation_matrix': 'one', 'ub_matrix': '1/Angstrom'}


def _fix_unit(name, value):
    if (unit := _matrix_units.get(name)) is not None:
        value.unit = unit
    return value


class NXsample(NXobject):
    """NXsample"""

    def __init__(self, attrs: dict[str, Any], children: dict[str, Field | Group]):
        super().__init__(attrs=attrs, children=children)
        for key in _matrix_units:
            if (field := children.get(key)) is not None:
                field.sizes = {k: field.sizes[k] for k in field.dims[:-2]}
                field.dtype = sc.DType.linear_transform3

    def read_children(self, sel: ScippIndex) -> sc.DataGroup:
        return sc.DataGroup(
            {
                name: _fix_unit(name, self.index_child(child, sel))
                for name, child in self._children.items()
            }
        )


base_definitions_dict['NXsample'] = NXsample



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import Any

import pytest
import scipp as sc

import scippnexus as snx

externalfile = pytest.importorskip('externalfile')

all_files = [
    '2023/DREAM_baseline_all_dets.nxs',
    '2023/BIFROST_873855_00000015.hdf',
    '2023/DREAM_mccode.h5',
    '2023/LOKI_mcstas_nexus_geometry.nxs',
    '2023/NMX_2e11-rechunk.h5',
    '2023/YMIR_038243_00010244.hdf',
    '2023/amor2020n000346_tweaked.nxs',
    '2023/LOKI_60322-2022-03-02_2205_fixed.nxs',
]


def get_item_at_path(dg: sc.DataGroup, path: str) -> sc.DataArray:
    nodes = path.split('/')
    for node in nodes:
        dg = dg[node]
    return dg


def assert_schema(dg: sc.DataGroup, schema: dict[str, Any]) -> None:
    for name, validate in schema.items():
        validate(get_item_at_path(dg, name))


def validator(
    item_type: type,
    sizes: dict[str, int] | None = None,
    dtype: str | sc.DType | None = None,
) -> None:
    def _validator(item):
        assert isinstance(item, item_type)
        if sizes is not None:
            assert item.sizes == sizes
        if dtype is not None:
            assert item.dtype == dtype

    return _validator


def bins_validator(item_type: type, sizes: dict[str, int] | None = None):
    def _validator(item):
        assert isinstance(item, item_type)
        assert item.bins is not None
        if sizes is not None:
            assert item.sizes == sizes

    return _validator


@pytest.mark.externalfile
@pytest.mark.parametrize('name', all_files)
def test_files_load_as_data_groups(name) -> None:
    with snx.File(externalfile.get_path(name)) as f:
        dg = f[()]
    assert isinstance(dg, sc.DataGroup)


@pytest.mark.externalfile
@pytest.mark.parametrize('name', all_files)
def test_files_load_as_data_groups_with_no_definitions(name) -> None:
    with snx.File(externalfile.get_path(name), definitions={}) as f:
        dg = f[()]
    assert isinstance(dg, sc.DataGroup)


@pytest.mark.externalfile
def test_amor2020n000346_tweaked() -> None:
    with snx.File(externalfile.get_path('2023/amor2020n000346_tweaked.nxs')) as f:
        dg = f[()]
    schema = {}
    schema['entry/instrument/multiblade_detector'] = bins_validator(
        sc.DataArray, {'detector_number': 9216}
    )
    schema['entry/stages/com'] = validator(sc.DataArray, {'time': 1, 'dim_1': 1})
    schema['entry/facility'] = validator(str)
    assert_schema(dg, schema)


@pytest.mark.externalfile
def test_LOKI_60322_2022_03_02_2205_fixed() -> None:
    with snx.File(
        externalfile.get_path('2023/LOKI_60322-2022-03-02_2205_fixed.nxs')
    ) as f:
        dg = f[()]
    schema = {}
    schema['entry/instrument/larmor_detector'] = bins_validator(
        sc.DataArray, {'detector_number': 458752}
    )
    schema['entry/instrument/monitor_1'] = bins_validator(
        sc.DataArray, {'event_time_zero': 1987}
    )
    schema['entry/instrument/monitor_2'] = bins_validator(
        sc.DataArray, {'event_time_zero': 1987}
    )
    schema['entry/instrument/source/transformations/trans_2'] = validator(
        sc.DataArray, {}, sc.DType.translation3
    )

    assert_schema(dg, schema)



import h5py
import numpy as np
import pytest
import scipp as sc
from scipp.testing import assert_identical

import scippnexus as snx
from scippnexus import NXdata, NXlog


@pytest.fixture
def h5root():
    """Yield h5py root group (file)"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        yield f


@pytest.fixture
def nxroot():
    """Yield NXroot containing a single NXentry named 'entry'"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        root = snx.Group(f, definitions=snx.base_definitions())
        root.create_class('entry', snx.NXentry)
        yield root


def test_without_coords(h5root) -> None:
    signal = sc.array(dims=['xx', 'yy'], unit='m', values=[[1.1, 2.2], [3.3, 4.4]])
    data = snx.create_class(h5root, 'data1', snx.NXdata)
    snx.create_field(data, 'signal', signal)
    data.attrs['axes'] = signal.dims
    data.attrs['signal'] = 'signal'
    obj = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(obj[...], sc.DataArray(signal))


def test_with_coords_matching_axis_names(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2, 3], [4, 5, 6]])
    )
    da.coords['xx'] = da.data['yy', 0]
    data = snx.create_class(h5root, 'data1', snx.NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'xx', da.coords['xx'])
    group = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(group[...], da)


def test_guessed_dim_for_coord_not_matching_axis_name(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2, 3], [4, 5, 6]])
    )
    da.coords['xx2'] = da.data['yy', 1]
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'xx2', da.coords['xx2'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], da)


def test_multiple_coords(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2, 3], [4, 5, 6]])
    )
    da.coords['xx'] = da.data['yy', 0]
    da.coords['xx2'] = da.data['yy', 1]
    da.coords['yy'] = da.data['xx', 0]
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'xx', da.coords['xx'])
    snx.create_field(data, 'xx2', da.coords['xx2'])
    snx.create_field(data, 'yy', da.coords['yy'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], da)


def test_slice_of_1d(h5root) -> None:
    da = sc.DataArray(sc.array(dims=['xx'], unit='m', values=[1, 2, 3]))
    da.coords['xx'] = da.data
    da.coords['xx2'] = da.data
    da.coords['scalar'] = sc.scalar(1.2)
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'xx', da.coords['xx'])
    snx.create_field(data, 'xx2', da.coords['xx2'])
    snx.create_field(data, 'scalar', da.coords['scalar'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data['xx', :2], da['xx', :2])
    assert sc.identical(data[:2], da['xx', :2])


def test_label_slice_of_1d(h5root) -> None:
    da = sc.DataArray(sc.array(dims=['xx'], unit='m', values=[1, 2, 3]))
    da.coords['xx'] = da.data
    da.coords['xx2'] = da.data
    da.coords['scalar'] = sc.scalar(1.2)
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'xx', da.coords['xx'])
    snx.create_field(data, 'xx2', da.coords['xx2'])
    snx.create_field(data, 'scalar', da.coords['scalar'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(
        data['xx', : sc.scalar(3, unit='m')], da['xx', : sc.scalar(3, unit='m')]
    )


def test_slice_of_multiple_coords(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2, 3], [4, 5, 6]])
    )
    da.coords['xx'] = da.data['yy', 0]
    da.coords['xx2'] = da.data['yy', 1]
    da.coords['yy'] = da.data['xx', 0]
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'xx', da.coords['xx'])
    snx.create_field(data, 'xx2', da.coords['xx2'])
    snx.create_field(data, 'yy', da.coords['yy'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(
        data['xx', : sc.scalar(3, unit='m')], da['xx', : sc.scalar(3, unit='m')]
    )


def test_guessed_dim_for_2d_coord_not_matching_axis_name(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2, 3], [4, 5, 6]])
    )
    da.coords['xx2'] = da.data
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'xx2', da.coords['xx2'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], da)


@pytest.mark.filterwarnings("ignore:Failed to load /data1:UserWarning")
def test_skips_axis_if_dim_guessing_finds_ambiguous_shape(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    )
    da.coords['yy2'] = da.data['xx', 0]
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'yy2', da.coords['yy2'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert set(data.dims) == {'dim_0', 'xx', 'yy'}
    dg = data[...]
    assert isinstance(dg, sc.DataGroup)
    assert 'yy2' in dg
    assert set(dg.dims) == {'dim_0', 'xx', 'yy'}


def test_dim_guessing_with_ambiguous_shape_accepts_multi_dim_match_at_end(
    h5root,
) -> None:
    da = sc.DataArray(
        sc.array(
            dims=['aux', 'xx', 'yy'],
            unit='m',
            values=[[[1, 2, 3], [4, 5, 6], [7, 8, 9]]],
        )
    )
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, '3x3', da.data['aux', 0])
    data = snx.Group(data, definitions=snx.base_definitions())
    loaded = data[...]
    assert_identical(loaded.data, da.data)
    assert_identical(loaded.coords['3x3'], da.data['aux', 0])


@pytest.mark.filterwarnings("ignore:Failed to load /data1:UserWarning")
def test_dim_guessing_with_ambiguous_shape_rejects_1d_dim_match_at_end(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    )
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, '1d', da.data['xx', 0])
    data = snx.Group(data, definitions=snx.base_definitions())
    dg = data[...]
    assert isinstance(dg, sc.DataGroup)
    assert '1d' in dg
    assert set(dg.dims) == {'dim_0', 'xx', 'yy'}


def test_guesses_transposed_dims_for_2d_coord(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2, 3], [4, 5, 6]])
    )
    da.coords['xx2'] = sc.transpose(da.data)
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'xx2', da.coords['xx2'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], da)


@pytest.mark.parametrize("indices", [1, [1]], ids=['int', 'list-of-int'])
def test_indices_attribute_for_coord(h5root, indices) -> None:
    da = sc.DataArray(sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2], [4, 5]]))
    da.coords['yy2'] = da.data['xx', 0]
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    data.attrs['yy2_indices'] = indices
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'yy2', da.coords['yy2'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], da)


@pytest.mark.parametrize("indices", [1, [1]], ids=['int', 'list-of-int'])
def test_indices_attribute_for_coord_with_nontrivial_slice(h5root, indices) -> None:
    da = sc.DataArray(sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2], [4, 5]]))
    da.coords['yy2'] = da.data['xx', 0]
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    data.attrs['yy2_indices'] = indices
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'yy2', da.coords['yy2'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data['yy', :1], da['yy', :1])


def test_transpose_indices_attribute_for_coord(h5root) -> None:
    da = sc.DataArray(sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2], [4, 5]]))
    da.coords['xx2'] = sc.transpose(da.data)
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    data.attrs['xx2_indices'] = [1, 0]
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'xx2', da.coords['xx2'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], da)


def test_auxiliary_signal_causes_load_as_dataset(h5root) -> None:
    signal = sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2, 3], [4, 5, 6]])
    aux = signal * 2
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = signal.dims
    data.attrs['signal'] = 'signal'
    # We flag 'xx' as auxiliary_signal. It should thus not be loaded as a coord,
    # even though we create the field.
    data.attrs['auxiliary_signals'] = ['xx']
    snx.create_field(data, 'signal', signal)
    snx.create_field(data, 'xx', aux)
    data = snx.Group(data, definitions=snx.base_definitions())
    assert_identical(data[...], sc.Dataset({'signal': signal, 'xx': aux}))


def test_NXlog_data_is_loaded_as_time_dependent_data_array(nxroot) -> None:
    da = sc.DataArray(
        data=sc.array(dims=['time'], unit='K', values=[1, 2, 3]),
        coords={
            'time': sc.epoch(unit='s')
            + sc.array(dims=['time'], unit='s', values=[1, 2, 3])
        },
    )
    data = nxroot.create_class('data1', NXdata)
    log = data.create_class('data', NXlog)
    log['time'] = da.coords['time']
    log['value'] = da.data
    data._group.attrs['signal'] = 'data'

    loaded = data[()]
    assert_identical(loaded, da)


@pytest.mark.parametrize("time_unit", ['m', None])
def test_NXlog_with_nontime_time_axis_can_be_loaded(nxroot, time_unit) -> None:
    da = sc.DataArray(
        data=sc.array(dims=['time'], unit='K', values=[1, 2, 3]),
        coords={'time': sc.array(dims=['time'], unit=time_unit, values=[1, 2, 3])},
    )
    data = nxroot.create_class('data1', NXdata)
    log = data.create_class('data', NXlog)
    log['time'] = da.coords['time']
    log['value'] = da.data
    loaded = data[()]
    assert_identical(loaded, da)
    assert loaded.coords['time'].dtype != sc.DType.datetime64


def test_field_dims_match_NXdata_dims(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2, 3], [4, 5, 6]])
    )
    da.coords['xx'] = da.data['yy', 0]
    da.coords['xx2'] = da.data['yy', 1]
    da.coords['yy'] = da.data['xx', 0]
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal1'
    snx.create_field(data, 'signal1', da.data)
    snx.create_field(data, 'xx', da.coords['xx'])
    snx.create_field(data, 'xx2', da.coords['xx2'])
    snx.create_field(data, 'yy', da.coords['yy'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data['xx', :2].data, data['signal1']['xx', :2])
    assert sc.identical(data['xx', :2].coords['xx'], data['xx']['xx', :2])
    assert sc.identical(data['xx', :2].coords['xx2'], data['xx2']['xx', :2])
    assert sc.identical(data['xx', :2].coords['yy'], data['yy'][:])


def test_field_dims_match_NXdata_dims_when_selected_via_class_name(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2, 3], [4, 5, 6]])
    )
    da.coords['xx'] = da.data['yy', 0]
    da.coords['xx2'] = da.data['yy', 1]
    da.coords['yy'] = da.data['xx', 0]
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal1'
    snx.create_field(data, 'signal1', da.data)
    snx.create_field(data, 'xx', da.coords['xx'])
    snx.create_field(data, 'xx2', da.coords['xx2'])
    snx.create_field(data, 'yy', da.coords['yy'])
    data = snx.Group(data, definitions=snx.base_definitions())
    fields = data[snx.Field]
    assert fields['signal1'].dims == ('xx', 'yy')
    assert fields['xx'].dims == ('xx',)
    assert fields['xx2'].dims == ('xx',)
    assert fields['yy'].dims == ('yy',)


@pytest.mark.filterwarnings("ignore:Failed to load /data1:UserWarning")
def test_uses_default_field_dims_if_inference_fails(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2, 3], [4, 5, 6]])
    )
    yy2 = sc.arange('yy', 5)
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'yy2', yy2)
    data = snx.Group(data, definitions=snx.base_definitions())
    dg = data[()]
    assert sc.identical(dg['yy2'], yy2.rename(yy='dim_0'))
    assert sc.identical(data['yy2'][()], yy2.rename(yy='dim_0'))


@pytest.mark.parametrize("unit", ['m', 's', None])
def test_create_field_from_variable(h5root, unit) -> None:
    var = sc.array(dims=['xx'], unit=unit, values=[3, 4])
    snx.create_field(h5root, 'field', var)
    group = snx.Group(h5root, definitions=snx.base_definitions())
    loaded = group['field'][...]
    # Nexus does not support storing dim labels
    assert sc.identical(loaded, var.rename(xx=loaded.dim))


def test_create_datetime_field_from_variable(h5root) -> None:
    var = sc.datetime(np.datetime64('now'), unit='ns') + sc.arange(
        'time', 1, 4, dtype='int64', unit='ns'
    )
    snx.create_field(h5root, 'field', var)
    group = snx.Group(h5root, definitions=snx.base_definitions())
    loaded = group['field'][...]
    # Nexus does not support storing dim labels
    assert sc.identical(loaded, var.rename(time=loaded.dim))


@pytest.mark.parametrize("nx_class", [NXdata, NXlog])
def test_create_class(nxroot, nx_class) -> None:
    group = nxroot.create_class('group', nx_class)
    assert group.nx_class == nx_class


def test_deprecated_errors_field_is_used_for_signal_errors(h5root) -> None:
    data = snx.create_class(h5root, 'data1', NXdata)
    values = sc.array(dims=['xx', 'yy'], unit='m', values=[[1.0, 2, 3], [4, 5, 6]])
    errors = sc.array(dims=['xx', 'yy'], unit='m', values=[[0.0, 2, 3], [0, 5, 6]])
    data.attrs['axes'] = values.dims
    data.attrs['signal'] = 'data'
    snx.create_field(data, 'data', values)
    snx.create_field(data, 'errors', errors)
    data = snx.Group(data, definitions=snx.base_definitions())
    loaded = data[()]
    assert_identical(sc.values(loaded), sc.DataArray(values))
    assert_identical(sc.stddevs(loaded), sc.DataArray(errors))


@pytest.mark.parametrize("errors_suffix", ['_error', '_errors'])
def test_field_matching_errors_regex_is_loaded_if_no_corresponding_value_field(
    h5root, errors_suffix
):
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2, 3], [4, 5, 6]])
    )
    da.coords[f'xx{errors_suffix}'] = da.data['yy', 0]
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, f'xx{errors_suffix}', da.coords[f'xx{errors_suffix}'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], da)


@pytest.mark.parametrize("errors_suffix", ['_error', '_errors'])
def test_uncertainties_of_coords_are_loaded(h5root, errors_suffix) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2, 3], [4, 5, 6]])
    )
    da.coords['xx'] = sc.array(
        dims=['xx'], unit='m', values=[1, 2, 3], variances=[1, 4, 9], dtype='float64'
    )
    da.coords['xx2'] = sc.array(
        dims=['xx'], unit='m', values=[2, 3], variances=[4, 9], dtype='float64'
    )
    da.coords['scalar'] = sc.scalar(value=1.2, variance=4.0, unit='K')
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    data.attrs['xx2_indices'] = 0
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'xx', sc.values(da.coords['xx']))
    snx.create_field(data, f'xx{errors_suffix}', sc.stddevs(da.coords['xx']))
    snx.create_field(data, 'xx2', sc.values(da.coords['xx2']))
    snx.create_field(data, f'xx2{errors_suffix}', sc.stddevs(da.coords['xx2']))
    snx.create_field(data, 'scalar', sc.values(da.coords['scalar']))
    snx.create_field(data, f'scalar{errors_suffix}', sc.stddevs(da.coords['scalar']))
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], da)


def test_unnamed_extra_dims_of_coords_are_squeezed(h5root) -> None:
    signal = sc.array(dims=['xx', 'yy'], unit='m', values=[[1.1, 2.2], [3.3, 4.4]])
    data = snx.create_class(h5root, 'data1', NXdata)
    snx.create_field(data, 'signal', signal)
    data.attrs['axes'] = signal.dims
    data.attrs['signal'] = 'signal'
    # shape=[1]
    snx.create_field(data, 'scalar', sc.array(dims=['ignored'], values=[1.2]))
    data = snx.Group(data, definitions=snx.base_definitions())
    loaded = data[...]
    assert sc.identical(loaded.coords['scalar'], sc.scalar(1.2))
    assert data['scalar'].ndim == 0
    assert data['scalar'].shape == ()
    assert sc.identical(data['scalar'][...], sc.scalar(1.2))


def test_unnamed_extra_dims_of_multidim_coords_are_squeezed(h5root) -> None:
    signal = sc.array(dims=['xx'], unit='m', values=[1.1, 2.2])
    data = snx.create_class(h5root, 'data1', NXdata)
    snx.create_field(data, 'signal', signal)
    data.attrs['axes'] = signal.dims
    data.attrs['signal'] = 'signal'
    # shape=[2,1]
    xx = sc.array(dims=['xx', 'ignored'], values=[[1.1], [2.2]])
    snx.create_field(data, 'xx', xx)
    data = snx.Group(data, definitions=snx.base_definitions())
    loaded = data[...]
    assert sc.identical(loaded.coords['xx'], xx['ignored', 0])
    assert data['xx'].ndim == 1
    assert data['xx'].shape == (2,)
    assert sc.identical(data['xx'][...], xx['ignored', 0])


def test_dims_of_length_1_are_kept_when_axes_specified(h5root) -> None:
    signal = sc.array(dims=['xx', 'yy'], unit='m', values=[[1.1]])
    data = snx.create_class(h5root, 'data1', NXdata)
    snx.create_field(data, 'signal', signal)
    data.attrs['axes'] = ['xx', 'yy']
    data.attrs['signal'] = 'signal'
    data = snx.Group(data, definitions=snx.base_definitions())
    loaded = data[...]
    assert sc.identical(loaded.data, signal)
    assert data['signal'].ndim == 2
    assert data['signal'].shape == (1, 1)


def test_only_dim_of_length_1_is_squeezed_when_no_axes_specified(h5root) -> None:
    signal = sc.array(dims=['xx'], unit='m', values=[1.1])
    data = snx.create_class(h5root, 'data1', NXdata)
    snx.create_field(data, 'signal', signal)
    data.attrs['signal'] = 'signal'
    data = snx.Group(data, definitions=snx.base_definitions())
    loaded = data[...]
    assert sc.identical(loaded.data, sc.scalar(1.1, unit='m'))
    assert data['signal'].ndim == 0
    assert data['signal'].shape == ()


def test_multi_dims_of_length_1_are_kept_when_no_axes_specified(h5root) -> None:
    signal = sc.array(dims=['xx', 'yy'], unit='m', values=[[1.1]])
    data = snx.create_class(h5root, 'data1', NXdata)
    snx.create_field(data, 'signal', signal)
    data.attrs['signal'] = 'signal'
    data = snx.Group(data, definitions=snx.base_definitions())
    loaded = data[...]
    assert sc.identical(
        loaded.data, sc.array(dims=['dim_0', 'dim_1'], unit='m', values=[[1.1]])
    )
    assert data['signal'].ndim == 2
    assert data['signal'].shape == (1, 1)


def test_one_dim_of_length_1_is_kept_when_no_axes_specified(h5root) -> None:
    signal = sc.array(dims=['xx', 'yy'], unit='m', values=[[1.1, 2.2]])
    data = snx.create_class(h5root, 'data1', NXdata)
    snx.create_field(data, 'signal', signal)
    data.attrs['signal'] = 'signal'
    data = snx.Group(data, definitions=snx.base_definitions())
    loaded = data[...]
    # Note that dimension gets renamed to `dim_0` since no axes are specified
    assert sc.identical(
        loaded.data, sc.array(dims=['dim_0', 'dim_1'], unit='m', values=[[1.1, 2.2]])
    )
    assert data['signal'].ndim == 2
    assert data['signal'].shape == (1, 2)
    assert data['signal'].dims == ('dim_0', 'dim_1')


def test_only_one_axis_specified_for_2d_field(h5root) -> None:
    signal = sc.array(dims=['xx', 'yy'], unit='m', values=[[1.1]])
    data = snx.create_class(h5root, 'data1', NXdata)
    snx.create_field(data, 'signal', signal)
    data.attrs['axes'] = ['zz']
    data.attrs['signal'] = 'signal'
    data = snx.Group(data, definitions=snx.base_definitions())
    loaded = data[...]
    assert sc.identical(loaded.data, sc.array(dims=['zz'], unit='m', values=[1.1]))


def test_fields_with_datetime_attribute_are_loaded_as_datetime(h5root) -> None:
    da = sc.DataArray(
        sc.epoch(unit='s')
        + sc.array(dims=['xx', 'yy'], unit='s', values=[[1, 2, 3], [4, 5, 6]])
    )
    da.coords['xx'] = da.data['yy', 0]
    da.coords['xx2'] = da.data['yy', 1]
    da.coords['yy'] = da.data['xx', 0]
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'xx', da.coords['xx'])
    snx.create_field(data, 'xx2', da.coords['xx2'])
    snx.create_field(data, 'yy', da.coords['yy'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], da)


def test_slicing_with_bin_edge_coord_returns_bin_edges(h5root) -> None:
    da = sc.DataArray(sc.array(dims=['xx'], unit='K', values=[1.1, 2.2, 3.3]))
    da.coords['xx'] = sc.array(dims=['xx'], unit='m', values=[0.1, 0.2, 0.3, 0.4])
    da.coords['xx2'] = sc.array(dims=['xx'], unit='m', values=[0.3, 0.4, 0.5, 0.6])
    data = snx.create_class(h5root, 'data', NXdata)
    snx.create_field(data, 'xx', da.coords['xx'])
    snx.create_field(data, 'xx2', da.coords['xx2'])
    snx.create_field(data, 'data', da.data)
    data.attrs['signal'] = 'data'
    data.attrs['axes'] = ['xx']
    data.attrs['xx_indices'] = [0]
    data.attrs['xx2_indices'] = [0]
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], da)
    assert sc.identical(data['xx', 0], da['xx', 0])
    assert sc.identical(data['xx', 1], da['xx', 1])
    assert sc.identical(data['xx', 0:1], da['xx', 0:1])
    assert sc.identical(data['xx', 1:3], da['xx', 1:3])
    assert sc.identical(data['xx', 1:1], da['xx', 1:1])  # empty slice


def test_legacy_signal_attr_is_used(h5root) -> None:
    signal = sc.array(dims=['xx', 'yy'], unit='m', values=[[1.1, 2.2], [3.3, 4.4]])
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = signal.dims
    field = snx.create_field(data, 'mysig', signal)
    field.attrs['signal'] = 1  # legacy way of defining signal
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], sc.DataArray(signal))


def test_invalid_group_signal_attribute_is_ignored(h5root) -> None:
    signal = sc.array(dims=['xx', 'yy'], unit='m', values=[[1.1, 2.2], [3.3, 4.4]])
    data = snx.create_class(h5root, 'data1', NXdata)
    data.attrs['axes'] = signal.dims
    data.attrs['signal'] = 'signal'
    field = snx.create_field(data, 'mysig', signal)
    field.attrs['signal'] = 1  # legacy way of defining signal
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], sc.DataArray(signal))


def test_legacy_axis_attrs_define_dim_names(h5root) -> None:
    da = sc.DataArray(sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2], [4, 5]]))
    da.coords['xx'] = da.data['yy', 0]
    da.coords['yy'] = da.data['xx', 0]
    data = snx.create_class(h5root, 'data1', NXdata)
    signal = snx.create_field(data, 'signal', da.data)
    xx = snx.create_field(data, 'xx', da.coords['xx'])
    yy = snx.create_field(data, 'yy', da.coords['yy'])
    signal.attrs['signal'] = 1
    xx.attrs['axis'] = 1
    yy.attrs['axis'] = 2
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], da)


@pytest.mark.parametrize('axis_sep', [':', ','])
def test_alternative_legacy_axis_attrs_with_signal_axes(h5root, axis_sep) -> None:
    da = sc.DataArray(sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2], [4, 5]]))
    da.coords['xx'] = da.data['yy', 0]
    da.coords['yy'] = da.data['xx', 0]
    data = snx.create_class(h5root, 'data1', NXdata)
    signal = snx.create_field(data, 'signal', da.data)
    xx = snx.create_field(data, 'xx', da.coords['xx'])
    yy = snx.create_field(data, 'yy', da.coords['yy'])
    signal.attrs['signal'] = 1
    # According to personal communication with Tobias Richter, the legacy 'axis'
    # attributes may have been used with two different meanings:
    # 1. As a 1-based integer (see other tests)
    # 2. As a boolean, where "1" means that the field is an axis.
    # In the latter case it appears to be useless since axis order cannot be inferred,
    # and we ignore the value if the signal has an `axes` attribute.
    signal.attrs['axes'] = axis_sep.join(['xx', 'yy'])
    xx.attrs['axis'] = 1
    yy.attrs['axis'] = 1
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], da)


def test_legacy_axis_favors_colon_sep(h5root) -> None:
    da = sc.DataArray(sc.array(dims=['xx,zz', 'yy'], unit='m', values=[[1, 2], [4, 5]]))
    da.coords['xx,zz'] = da.data['yy', 0]
    da.coords['yy'] = da.data['xx,zz', 0]
    data = snx.create_class(h5root, 'data1', NXdata)
    signal = snx.create_field(data, 'signal', da.data)
    xx = snx.create_field(data, 'xx,zz', da.coords['xx,zz'])
    yy = snx.create_field(data, 'yy', da.coords['yy'])
    signal.attrs['signal'] = 1
    signal.attrs['axes'] = 'xx,zz:yy'
    xx.attrs['axis'] = 1
    yy.attrs['axis'] = 1
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], da)


@pytest.mark.parametrize('axis_sep', [':', ','])
def test_guesses_dims_of_bin_edge_fields(h5root, axis_sep) -> None:
    da = sc.DataArray(sc.ones(dims=['xx', 'yy'], unit='m', shape=(2, 4)))
    da.coords['xx2'] = sc.array(dims=['xx'], unit='m', values=[1, 2, 3])
    da.coords['yy2'] = sc.array(dims=['yy'], unit='m', values=[1, 2, 3, 4, 5])
    data = snx.create_class(h5root, 'data1', NXdata)
    signal = snx.create_field(data, 'signal', da.data)
    signal.attrs['signal'] = 1
    signal.attrs['axes'] = axis_sep.join(['xx', 'yy'])
    snx.create_field(data, 'xx2', da.coords['xx2'])
    snx.create_field(data, 'yy2', da.coords['yy2'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], da)


@pytest.mark.parametrize('axis_sep', [':', ','])
def test_guesses_dims_of_2d_bin_edge_fields(h5root, axis_sep) -> None:
    da = sc.DataArray(sc.ones(dims=['xx', 'yy'], unit='m', shape=(2, 4)))
    da.coords['xx2'] = sc.ones(dims=['xx', 'yy'], unit='m', shape=(2, 5))
    data = snx.create_class(h5root, 'data1', NXdata)
    signal = snx.create_field(data, 'signal', da.data)
    signal.attrs['signal'] = 1
    signal.attrs['axes'] = axis_sep.join(['xx', 'yy'])
    snx.create_field(data, 'xx2', da.coords['xx2'])
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], da)


@pytest.mark.filterwarnings("ignore:Failed to load /data1:UserWarning")
def test_nested_groups_trigger_fallback_to_load_as_data_group(h5root) -> None:
    da = sc.DataArray(sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2], [4, 5]]))
    data = snx.create_class(h5root, 'data1', NXdata)
    snx.create_field(data, 'signal', da.data)
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    snx.create_class(data, 'nested', NXdata)
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], sc.DataGroup(signal=da.data, nested=sc.DataGroup()))


def test_slicing_raises_given_invalid_index(h5root) -> None:
    signal = sc.array(dims=['xx', 'yy'], unit='m', values=[[1.1, 2.2], [3.3, 4.4]])
    data = snx.create_class(h5root, 'data1', NXdata)
    snx.create_field(data, 'signal', signal)
    data.attrs['axes'] = signal.dims
    data.attrs['signal'] = 'signal'
    data = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(data[...], sc.DataArray(signal))
    with pytest.raises(IndexError):
        data['xx', 2]
    with pytest.raises(sc.DimensionError):
        data['zz', 0]


def test_label_based_slicing(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['xx', 'yy'], unit='m', values=[[1, 2], [4, 5]]),
        coords={
            'xx': sc.array(dims=['xx'], unit='m', values=[1.0, 2.0]),
            'yy': sc.array(dims=['yy'], unit='m', values=[0.1, 0.0]),
        },
    )
    data = snx.create_class(h5root, 'data1', NXdata)
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'xx', da.coords['xx'])
    snx.create_field(data, 'yy', da.coords['yy'])
    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    data = snx.Group(data, definitions=snx.base_definitions())
    sc.testing.assert_identical(
        data['xx', sc.scalar(1.0, unit='m') : sc.scalar(3.0, unit='m')],
        da['xx', sc.scalar(1.0, unit='m') : sc.scalar(3.0, unit='m')],
    )
    sc.testing.assert_identical(
        data['yy', sc.scalar(0.2, unit='m') : sc.scalar(0.01, unit='m')],
        da['yy', sc.scalar(0.2, unit='m') : sc.scalar(0.01, unit='m')],
    )


def create_nxdata(h5root, dims, coords):
    da = sc.DataArray(
        sc.array(dims=dims, unit='m', values=np.random.randn(*(5 for _ in dims))),
        coords={
            coord: sc.linspace(dim, 0, 1, 5, unit='s')
            for dim, coord in zip(dims, coords, strict=True)
        },
    )
    data = snx.create_class(h5root, 'data1', NXdata)
    snx.create_field(data, 'signal', da.data)
    for dim, coord in zip(dims, coords, strict=True):
        snx.create_field(data, dim, da.coords[coord])

    data.attrs['axes'] = da.dims
    data.attrs['signal'] = 'signal'
    data = snx.Group(data, definitions=snx.base_definitions())
    return data


@pytest.mark.parametrize(
    "slice_",
    [
        ('time', sc.scalar(1, unit='s')),
        ('time', slice(sc.scalar(1, unit='s'), None)),
        ('time', slice(None, sc.scalar(1, unit='s'))),
        ('time', slice(sc.scalar(1, unit='s'), sc.scalar(3, unit='s'))),
        {'time': sc.scalar(1, unit='s'), 'x': sc.scalar(1, unit='s')},
        {'x': sc.scalar(1, unit='s')},
    ],
)
@pytest.mark.parametrize(
    ('dims', 'coords'),
    [
        (('time',), ('time',)),
        (('time',), ('time2',)),
        (('time', 'x'), ('time2', 'x')),
        (('x', 'y'), ('time2', 'time')),
        (('x', 'y'), ('x', 'y')),
    ],
)
def test_label_indexing_dataset_behaves_same_as_indexing_scipp_dataarray(
    h5root, slice_, dims, coords
):
    nx = create_nxdata(h5root, dims, coords)
    da = nx[()]

    exception = None
    try:
        # Scipp does not support dict slicing,
        # manually slice datagroup in multiple coords
        if isinstance(slice_, dict):
            for s in slice_.items():
                da = da[s]
        else:
            da = da[slice_]
    except Exception as e:
        exception = type(e)

    if exception:
        with pytest.raises(exception):
            nx[slice_]
    else:
        assert_identical(nx[slice_], da)


def test_scalar_signal_without_unit_works(h5root) -> None:
    da = sc.DataArray(
        sc.scalar(1.1, unit=None), coords={'xx': sc.scalar(2.0, unit='m')}
    )
    data = snx.create_class(h5root, 'data1', snx.NXdata)
    data.attrs['axes'] = []
    data.attrs['signal'] = 'signal'
    snx.create_field(data, 'signal', da.data)
    snx.create_field(data, 'xx', da.coords['xx'])
    group = snx.Group(data, definitions=snx.base_definitions())
    assert sc.identical(group[...], da)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from typing import Any, ClassVar

import scipp as sc

from .base import Group, NexusStructureError, NXobject, base_definitions_dict
from .field import Field


def _parse(
    *,
    vertices: sc.Variable,
    cylinders: sc.Variable,
    detector_number: sc.Variable | None = None,
    parent_detector_number: sc.Variable | None = None,
) -> sc.Variable:
    face1_center = cylinders['vertex_index', 0]
    face1_edge = cylinders['vertex_index', 1]
    face2_center = cylinders['vertex_index', 2]
    ds = sc.Dataset(
        {
            'face1_center': vertices[face1_center.values],
            'face1_edge': vertices[face1_edge.values],
            'face2_center': vertices[face2_center.values],
        }
    )
    ds = ds.rename(**{vertices.dim: 'cylinder'})
    if detector_number is None:
        # All cylinders belong to the same shape
        return sc.bins(begin=sc.index(0), dim='cylinder', data=ds)
    if parent_detector_number is None:
        raise NexusStructureError(
            "`detector_number` not given, but "
            "NXcylindrical_geometry contains mapping to `detector_number`."
        )
    # detector_number gives indices into cylinders, the naming in the NeXus
    # standard appears to be misleading
    if parent_detector_number.values.size != detector_number.values.size:
        raise NexusStructureError(
            "Number of detector numbers in NXcylindrical_geometry "
            "does not match the one given by the parent."
        )
    detecting_cylinders = ds['cylinder', detector_number.values]
    # One cylinder per detector
    begin = sc.arange(
        'dummy', parent_detector_number.values.size, unit=None, dtype='int64'
    )
    end = begin + sc.index(1)
    shape = sc.bins(begin=begin, end=end, dim='cylinder', data=detecting_cylinders)
    return shape.fold(dim='dummy', sizes=parent_detector_number.sizes)


class NXcylindrical_geometry(NXobject):
    _dims: ClassVar[dict[str, tuple[str, ...]]] = {
        'vertices': ('vertex',),
        'detector_number': ('detector_number',),
        'cylinders': ('cylinder', 'vertex_index'),
    }

    def __init__(self, attrs: dict[str, Any], children: dict[str, Field | Group]):
        super().__init__(attrs=attrs, children=children)
        for name, field in children.items():
            if isinstance(field, Field):
                if name == 'vertices':
                    # Shape has one more entry than dims, the vector dim.
                    field.sizes = dict(
                        zip(self._dims.get(name), field.dataset.shape[:-1], strict=True)
                    )
                    field.dtype = sc.DType.vector3
                else:
                    field.sizes = dict(
                        zip(self._dims.get(name), field.dataset.shape, strict=True)
                    )

    def load_as_array(self, detector_number: sc.Variable | None = None) -> sc.Variable:
        return _parse(**self[()], parent_detector_number=detector_number)

    @staticmethod
    def assemble_as_child(
        children: sc.DataGroup, detector_number: sc.Variable | None = None
    ) -> sc.Variable:
        return _parse(**children, parent_detector_number=detector_number)


base_definitions_dict['NXcylindrical_geometry'] = NXcylindrical_geometry



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from typing import Any

_version = '1'

__all__ = ['get_path']


def _make_pooch() -> Any:
    import pooch

    return pooch.create(
        path=pooch.os_cache('scippnexus'),
        env='SCIPPNEXUS_DATA_DIR',
        retry_if_failed=3,
        base_url='https://public.esss.dk/groups/scipp/scippnexus/{version}/',
        version=_version,
        registry={
            'PG3_4844_event.nxs': 'md5:d5ae38871d0a09a28ae01f85d969de1e',
        },
    )


_pooch = _make_pooch()


def get_path(name: str) -> str:
    """
    Return the path to a data file bundled with scippnexus.

    This function only works with example data and cannot handle
    paths to custom files.
    """
    return _pooch.fetch(name)  # type: ignore[no-any-return]



import pytest
import scipp as sc

from scippnexus import DependsOn
from scippnexus.nxtransformations import Transform, TransformationError


@pytest.fixture
def depends_on() -> DependsOn:
    return DependsOn(parent='/', value='.')


@pytest.fixture
def z_vector() -> sc.Variable:
    return sc.vector(value=[0, 0, 1], unit='m')


def test_init_raises_if_transformation_type_is_invalid(depends_on, z_vector) -> None:
    with pytest.raises(TransformationError, match='transformation_type'):
        Transform(
            name='t1',
            transformation_type='trans',
            value=sc.ones(dims=['x', 'y', 'z'], shape=(2, 3, 4)),
            vector=z_vector,
            depends_on=depends_on,
        )


def test_sizes_returns_sizes_of_value(depends_on, z_vector) -> None:
    value = sc.ones(dims=['x', 'y', 'z'], shape=(2, 3, 4))
    transform = Transform(
        name='t1',
        transformation_type='translation',
        value=value,
        vector=z_vector,
        depends_on=depends_on,
    )
    assert transform.sizes == {'x': 2, 'y': 3, 'z': 4}



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import h5py
import numpy as np
import pytest
import scipp as sc
from scipp.testing import assert_identical

import scippnexus as snx
from scippnexus import NXentry, NXlog

# representative sample of UTF-8 test strings from
# https://www.w3.org/2001/06/utf-8-test/UTF-8-demo.html
UTF8_TEST_STRINGS = (
    "∮ E⋅da = Q,  n → ∞, ∑ f(i) = ∏ g(i), ∀x∈ℝ: ⌈x⌉ = −⌊−x⌋, α ∧ ¬β = ¬(¬α ∨ β)",  # noqa: RUF001
    "2H₂ + O₂ ⇌ 2H₂O, R = 4.7 kΩ, ⌀ 200 mm",
    "Σὲ γνωρίζω ἀπὸ τὴν κόψη",
)


@pytest.fixture
def h5root():
    """Yield h5py root group (file)"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        yield f


@pytest.fixture
def nxroot():
    """Yield NXroot containing a single NXentry named 'entry'"""
    with h5py.File('dummy.nxs', mode='w', driver="core", backing_store=False) as f:
        root = snx.Group(f, definitions=snx.base_definitions())
        root.create_class('entry', NXentry)
        yield root


def test_nxobject_log(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['time'], values=[1.1, 2.2, 3.3]),
        coords={
            'time': sc.epoch(unit='ns')
            + sc.array(dims=['time'], unit='s', values=[4.4, 5.5, 6.6]).to(
                unit='ns', dtype='int64'
            )
        },
    )
    log = snx.create_class(h5root, 'log', NXlog)
    snx.create_field(log, 'value', da.data)
    snx.create_field(log, 'time', da.coords['time'] - sc.epoch(unit='ns'))
    log = snx.Group(log, definitions=snx.base_definitions())
    assert sc.identical(log[...], da)


def test_nxlog_with_missing_value_uses_time_as_value(nxroot) -> None:
    time = sc.epoch(unit='ns') + sc.array(
        dims=['time'], unit='s', values=[4.4, 5.5, 6.6]
    ).to(unit='ns', dtype='int64')
    log = nxroot['entry'].create_class('log', NXlog)
    log['time'] = time - sc.epoch(unit='ns')
    loaded = log[()]
    assert_identical(loaded, sc.DataArray(data=time))


def test_nxlog_length_1(h5root) -> None:
    nxroot = snx.Group(h5root, definitions=snx.base_definitions())
    da = sc.DataArray(
        sc.array(dims=['time'], values=[1.1]),
        coords={
            'time': sc.epoch(unit='ns')
            + sc.array(dims=['time'], unit='s', values=[4.4]).to(
                unit='ns', dtype='int64'
            )
        },
    )
    log = nxroot.create_class('log', NXlog)
    log['value'] = da.data
    log['time'] = da.coords['time'] - sc.epoch(unit='ns')
    assert sc.identical(log[...], da)


def test_nxlog_length_1_two_dims_no_time_defaults_inner_dim_name(nxroot) -> None:
    var = sc.array(dims=['time', 'ignored'], values=[[1.1]])
    log = nxroot['entry'].create_class('log', NXlog)
    log['value'] = var
    assert_identical(log[...], sc.DataArray(var.rename(ignored='dim_1')))


def test_nxlog_length_1_two_dims_with_time_defaults_inner_dim_name(nxroot) -> None:
    da = sc.DataArray(
        sc.array(dims=['time', 'ignored'], values=[[1.1]]),
        coords={
            'time': sc.epoch(unit='ns')
            + sc.array(dims=['time'], unit='s', values=[4.4]).to(
                unit='ns', dtype='int64'
            )
        },
    )
    log = nxroot['entry'].create_class('log', NXlog)
    log['value'] = da.data
    log['time'] = da.coords['time'] - sc.epoch(unit='ns')
    assert sc.identical(log[...], da.rename(ignored='dim_1'))


def test_nxlog_axes_replaces_time_dim(nxroot) -> None:
    da = sc.DataArray(
        sc.array(dims=['time', 'ignored'], values=[[1.1]]),
        coords={
            'time': sc.epoch(unit='ns')
            + sc.array(dims=['time'], unit='s', values=[4.4]).to(
                unit='ns', dtype='int64'
            )
        },
    )
    log = nxroot['entry'].create_class('log', NXlog)
    log._group.attrs['axes'] = ['yy', 'xx']
    log['value'] = da.data
    log['time'] = da.coords['time'] - sc.epoch(unit='ns')
    expected = sc.DataArray(
        sc.array(dims=['yy', 'xx'], values=[[1.1]]),
        coords={'time': da.coords['time'].squeeze()},
    )
    assert sc.identical(log[...], expected)


def test_nxlog_three_dims_with_time_of_length_1(nxroot) -> None:
    da = sc.DataArray(
        sc.array(dims=['time', 'a', 'b'], values=np.arange(9.0).reshape(1, 3, 3)),
        coords={
            'time': sc.epoch(unit='ns')
            + sc.array(dims=['time'], unit='s', values=[4.4]).to(
                unit='ns', dtype='int64'
            )
        },
    )
    log = nxroot['entry'].create_class('log', NXlog)
    log['value'] = da.data
    log['time'] = da.coords['time'] - sc.epoch(unit='ns')
    loaded = log[...]
    assert_identical(
        loaded.data,
        sc.array(
            dims=['time', 'dim_1', 'dim_2'], values=np.arange(9.0).reshape(1, 3, 3)
        ),
    )


def test_nxlog_with_shape_0(nxroot) -> None:
    da = sc.DataArray(
        sc.ones(dims=['time', 'ignored'], shape=(0, 1)),
        coords={'time': sc.ones(dims=['time'], shape=(0,), unit='s')},
    )
    log = nxroot['entry'].create_class('log', NXlog)
    log['value'] = da.data
    log['time'] = da.coords['time']
    da.coords['time'] = sc.datetimes(dims=['time'], values=[], unit='ns')
    assert_identical(log[...], da.rename(ignored='dim_1'))


def test_log_with_connection_status_raises_with_positional_and_label_indexing(
    h5root,
) -> None:
    da = sc.DataArray(
        sc.array(dims=['time'], values=[1.1, 2.2, 3.3]),
        coords={
            'time': sc.array(
                dims=['time'], unit='s', values=[44, 55, 66], dtype='int64'
            )
        },
    )
    log = snx.create_class(h5root, 'log', NXlog)
    snx.create_field(log, 'value', da.data)
    snx.create_field(log, 'time', da.coords['time'])
    connection_status = da['time', :2].copy()
    snx.create_field(log, 'connection_status', connection_status.data)
    snx.create_field(log, 'connection_status_time', connection_status.coords['time'])
    log = snx.Group(log, definitions=snx.base_definitions())
    with pytest.raises(sc.DimensionError):
        log['time', :2]
    with pytest.raises(sc.DimensionError):
        log['time', : sc.scalar(60, unit='s')]
    with pytest.raises(sc.DimensionError):
        log[:2]
    with pytest.raises(sc.DimensionError):
        log[:]


@pytest.mark.parametrize('sublog_length', [0, 1, 2])
def test_log_with_connection_status_loaded_as_datagroup_containing_data_arrays(
    h5root, sublog_length
):
    da = sc.DataArray(
        sc.array(dims=['time'], values=[1.1, 2.2, 3.3]),
        coords={
            'time': sc.array(
                dims=['time'], unit='s', values=[44, 55, 66], dtype='int64'
            )
        },
    )
    log = snx.create_class(h5root, 'log', NXlog)
    snx.create_field(log, 'value', da.data)
    snx.create_field(log, 'time', da.coords['time'])
    connection_status = da['time', :sublog_length].copy()
    snx.create_field(log, 'connection_status', connection_status.data)
    snx.create_field(log, 'connection_status_time', connection_status.coords['time'])
    log = snx.Group(log, definitions=snx.base_definitions())
    loaded = log[()]
    da.coords['time'] = sc.epoch(unit='s') + da.coords['time']
    connection_status.coords['time'] = (
        sc.epoch(unit='s') + connection_status.coords['time']
    )
    assert_identical(loaded['value'], da)
    assert_identical(loaded['connection_status'], connection_status)


def test_log_with_alarm_loaded_as_datagroup_containing_data_arrays(h5root) -> None:
    da = sc.DataArray(
        sc.array(dims=['time'], values=[1.1, 2.2, 3.3]),
        coords={
            'time': sc.array(
                dims=['time'], unit='s', values=[44, 55, 66], dtype='int64'
            )
        },
    )
    log = snx.create_class(h5root, 'log', NXlog)
    snx.create_field(log, 'value', da.data)
    snx.create_field(log, 'time', da.coords['time'])
    alarm = da['time', :2].copy()
    alarm.coords['message'] = sc.array(dims=['time'], values=['alarm 1', 'alarm 2'])
    snx.create_field(log, 'alarm_severity', alarm.data)
    snx.create_field(log, 'alarm_message', alarm.coords['message'])
    snx.create_field(log, 'alarm_time', alarm.coords['time'])
    log = snx.Group(log, definitions=snx.base_definitions())
    loaded = log[()]
    da.coords['time'] = sc.epoch(unit='s') + da.coords['time']
    alarm.coords['time'] = sc.epoch(unit='s') + alarm.coords['time']
    assert_identical(loaded['value'], da)
    assert_identical(loaded['alarm'], alarm)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from __future__ import annotations

import types
from collections.abc import Callable, KeysView, Mapping
from typing import Any, Protocol, TypeAlias

import numpy.typing as npt


class H5Base(Protocol):
    @property
    def attrs(self) -> dict[str, object]:  # TODO better type hint
        """Attributes of dataset or group"""

    @property
    def name(self) -> str:
        """Name of dataset or group"""

    @property
    def file(self) -> Any:
        """File of dataset or group"""

    @property
    def parent(self) -> H5Group:
        """Parent of dataset or group"""


# TODO Define more required methods
class H5Dataset(H5Base, Protocol):
    """h5py.Dataset-like"""

    @property
    def shape(self) -> tuple[int, ...]:
        """Shape of a dataset"""

    @property
    def dtype(self) -> list[int]:
        """dtype of a dataset"""

    def read_direct(self, array: npt.NDArray[Any]) -> None:
        """Read dataset into given buffer"""


class H5Group(H5Base, Protocol):
    """h5py.Group-like"""

    def __getitem__(self, index: str | Any) -> H5Dataset | H5Group:
        """Keys in the group"""

    def keys(self) -> KeysView[str]:
        """Keys in the group"""

    def create_dataset(self) -> H5Dataset:
        """Create a dataset"""

    def create_group(self) -> H5Group:
        """Create a group"""

    def visititems(
        self,
        func: Callable[[str], None],
    ) -> None:
        """Apply callable to all items, recursively"""


# Note that scipp does not support dicts yet, but this HDF5 code does, to
# allow for loading blocks of 2d (or higher) data efficiently.
ScippIndex: TypeAlias = (
    types.EllipsisType
    | int
    | tuple[()]
    | slice
    | tuple[str, int | slice]
    | dict[str, int | slice]
)

Definitions = Mapping[str, type]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
import io
import os
from contextlib import AbstractContextManager
from typing import Any

import h5py

from .base import (
    DefaultDefinitions,
    DefaultDefinitionsType,
    Group,
    base_definitions,
)
from .typing import Definitions


class File(AbstractContextManager[Group], Group):
    def __init__(
        self,
        name: str | os.PathLike[str] | io.BytesIO | h5py.Group,
        *args: Any,
        definitions: Definitions | DefaultDefinitionsType = DefaultDefinitions,
        **kwargs: Any,
    ) -> None:
        """Context manager for NeXus files, similar to h5py.File.

        Arguments other than documented are as in :py:class:`h5py.File`.

        Parameters
        ----------
        name:
            Specifies the file to open.
            If this is a :class:`hyp5.File` object, the `:class:`File` will wrap
            this file handle but will not close it when used as a context manager.
        definitions:
            Mapping of NX_class names to application-specific definitions.
            The default is to use the base definitions as defined in the
            NeXus standard.
        """
        defs: Definitions = (
            base_definitions() if definitions is DefaultDefinitions else definitions  # type: ignore[assignment]
        )

        if isinstance(name, h5py.File | h5py.Group):
            if args or kwargs:
                raise TypeError('Cannot provide both h5py.File and other arguments')
            self._file = name
            self._manage_file_context = False
        else:
            self._file = h5py.File(name, *args, **kwargs)
            self._manage_file_context = True
        super().__init__(self._file, definitions=defs)

    def __enter__(self) -> Group:
        if self._manage_file_context:
            self._file.__enter__()
        return self

    def __exit__(self, exc_type: object, exc_value: object, traceback: object) -> None:
        if self._manage_file_context:
            self._file.close()

    def close(self) -> None:
        self._file.close()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
from __future__ import annotations

import datetime
import posixpath
import re
import warnings
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import h5py as h5
import numpy as np
import scipp as sc

from scippnexus._common import convert_time_to_datetime64, to_plain_index
from scippnexus._string import warn_latin1_decode
from scippnexus.typing import H5Dataset, ScippIndex

from ._cache import cached_property
from ._common import has_time_unit
from .attrs import Attrs

if TYPE_CHECKING:
    from .base import Group


@dataclass
class DependsOn:
    """
    Represents a depends_on reference in a NeXus file.

    The parent (the full path within the NeXus file) is stored, as the value may be
    relative or absolute, so having the path available after loading is essential.
    """

    parent: str
    value: str

    def absolute_path(self) -> str | None:
        if self.is_terminal:
            return None
        return posixpath.normpath(posixpath.join(self.parent, self.value))

    @property
    def is_terminal(self) -> bool:
        return self.value == '.'


def _as_datetime(obj: Any) -> sc.Variable | None:
    if isinstance(obj, str):
        try:
            # NumPy and scipp cannot handle timezone information. We therefore apply it,
            # i.e., convert to UTC.
            # Would like to use datetime directly, but with Python's datetime we do not
            # get nanosecond precision. Therefore we combine numpy and datetime parsing.
            date_only = 'T' not in obj
            if date_only:
                return sc.datetime(obj)
            _, time = obj.split('T')
            time_and_timezone_offset = re.split(r'Z|\+|-', time)
            time = time_and_timezone_offset[0]
            if len(time_and_timezone_offset) == 1:
                # No timezone, parse directly (scipp based on numpy)
                return sc.datetime(obj)
            else:
                # There is timezone info. Parse with datetime.
                dt = (
                    datetime.datetime.fromisoformat(obj)
                    .replace(microsecond=0)  # handled by numpy
                    .astimezone(datetime.UTC)
                    .replace(tzinfo=None)
                    .isoformat()
                )
                # We operate with string operations here and thus end up parsing date
                # and time twice. The reason is that the timezone-offset arithmetic
                # cannot be done, e.g., in nanoseconds without causing rounding errors.
                if '.' in time:
                    dt += f".{time.split('.')[1]}"
                return sc.datetime(dt)
        except ValueError:
            pass
    return None


_scipp_dtype = {
    np.dtype('int8'): sc.DType.int32,
    np.dtype('int16'): sc.DType.int32,
    np.dtype('uint8'): sc.DType.int32,
    np.dtype('uint16'): sc.DType.int32,
    np.dtype('uint32'): sc.DType.int32,
    np.dtype('uint64'): sc.DType.int64,
    np.dtype('int32'): sc.DType.int32,
    np.dtype('int64'): sc.DType.int64,
    np.dtype('float32'): sc.DType.float32,
    np.dtype('float64'): sc.DType.float64,
    np.dtype('bool'): sc.DType.bool,
}


def _dtype_fromdataset(dataset: H5Dataset) -> sc.DType:
    return _scipp_dtype.get(dataset.dtype, sc.DType.string)


def _squeezed_field_sizes(dataset: H5Dataset) -> dict[str, int]:
    if (shape := dataset.shape) == (1,):
        return {}
    return {f'dim_{i}': size for i, size in enumerate(shape)}


@dataclass(init=False)
class Field:
    """NeXus field.

    In HDF5 fields are represented as datasets.
    """

    dataset: h5.Dataset
    parent: Group
    sizes: dict[str, int]
    dtype: sc.DType
    errors: h5.Dataset | None

    def __init__(
        self,
        dataset: h5.Dataset,
        parent: Group,
        sizes: dict[str, int] | None = None,
        dtype: sc.DType | None = None,
        errors: h5.Dataset | None = None,
    ) -> None:
        self.dataset = dataset
        self.parent = parent
        self.sizes = sizes if sizes is not None else _squeezed_field_sizes(dataset)
        self.dtype = dtype if dtype is not None else _dtype_fromdataset(dataset)
        self.errors = errors

    @cached_property
    def attrs(self) -> Attrs:
        """The attributes of the dataset.

        Cannot be used for writing attributes, since they are cached for performance.
        """
        return Attrs(self.dataset.attrs)

    @property
    def dims(self) -> tuple[str, ...]:
        return tuple(self.sizes.keys())

    @property
    def shape(self) -> tuple[int, ...]:
        return tuple(self.sizes.values())

    @cached_property
    def file(self) -> Group:
        return self.parent.file

    def __getitem__(self, select: ScippIndex) -> Any | sc.Variable:
        """
        Load the field as a :py:class:`scipp.Variable` or Python object.

        If the shape is empty and no unit is given this returns a Python object, such
        as a string or integer. Otherwise a :py:class:`scipp.Variable` is returned.

        Parameters
        ----------
        select:
            Scipp-style index: Load the specified slice of the current group.

        Returns
        -------
        :
            Loaded data.
        """
        from .nxtransformations import maybe_transformation

        index = to_plain_index(self.dims, select)
        if isinstance(index, int | slice):
            index = (index,)

        base_dims = self.dims
        base_shape = self.shape
        dims = []
        shape = []
        for i, ind in enumerate(index):
            if not isinstance(ind, int):
                dims.append(base_dims[i])
                shape.append(len(range(*ind.indices(base_shape[i]))))

        variable = sc.empty(
            dims=dims,
            shape=shape,
            dtype=self.dtype,
            unit=self.unit,
            with_variances=self.errors is not None,
        )

        # If the variable is empty, return early
        if np.prod(shape) == 0:
            variable = self._maybe_datetime(variable)
            return maybe_transformation(self, value=variable)

        if self.dtype == sc.DType.string:
            try:
                strings = self.dataset.asstr()[index]
            except UnicodeDecodeError as e:
                strings = self.dataset.asstr(encoding='latin-1')[index]
                warn_latin1_decode(self.dataset, strings, e)
            variable.values = np.asarray(strings).flatten()
            if self.dataset.name.endswith('/depends_on') and variable.ndim == 0:
                return DependsOn(parent=self.dataset.parent.name, value=variable.value)
        elif variable.values.flags["C_CONTIGUOUS"]:
            # On versions of h5py prior to 3.2, a TypeError occurs in some cases
            # where h5py cannot broadcast data with e.g. shape (20, 1) to a buffer
            # of shape (20,). Note that broadcasting (1, 20) -> (20,) does work
            # (see https://github.com/h5py/h5py/pull/1796).
            # Therefore, we manually squeeze here.
            # A pin of h5py<3.2 is currently required by Mantid and hence scippneutron
            # (see https://github.com/h5py/h5py/issues/1880#issuecomment-823223154)
            # hence this workaround. Once we can use a more recent h5py with Mantid,
            # this try/except can be removed.
            try:
                self.dataset.read_direct(variable.values, source_sel=index)
            except TypeError:
                variable.values = self.dataset[index].squeeze()
            if self.errors is not None:
                _load_variances(self.errors, variable, index)
        else:
            variable.values = self.dataset[index]
        if variable.ndim == 0 and variable.unit is None and variable.fields is None:
            # Work around scipp/scipp#2815, and avoid returning NumPy bool
            if isinstance(variable.values, np.ndarray) and variable.dtype != 'bool':
                return variable.values[()]
            else:
                return variable.value
        variable = self._maybe_datetime(variable)
        return maybe_transformation(self, value=variable)

    def _maybe_datetime(self, variable: sc.Variable) -> sc.Variable:
        if has_time_unit(variable):
            starts = [
                dt
                for name in self.attrs
                if (dt := _as_datetime(self.attrs[name])) is not None
            ]
            if len(starts) == 1:
                variable = convert_time_to_datetime64(
                    variable,
                    start=starts[0],
                    scaling_factor=self.attrs.get('scaling_factor'),
                )

        return variable

    def __repr__(self) -> str:
        return f'<Nexus field "{self.dataset.name}">'

    @property
    def name(self) -> str:
        return self.dataset.name  # type: ignore[no-any-return]

    @property
    def ndim(self) -> int:
        """Total number of dimensions in the dataset.
        See the shape property for potential differences to the value returned by the
        underlying h5py.Dataset.ndim.
        """
        return len(self.shape)

    @cached_property
    def unit(self) -> sc.Unit | None:
        if (unit := self.attrs.get('units')) is not None:
            try:
                return sc.Unit(unit)
            except sc.UnitError:
                warnings.warn(
                    f"Unrecognized unit '{unit}' for value dataset "
                    f"in '{self.name}'; setting unit as 'dimensionless'",
                    stacklevel=2,
                )
                return sc.units.one  # type: ignore[no-any-return]
        return None


def _load_variances(
    dataset: h5.Dataset, var: sc.Variable, index: int | slice | tuple[int | slice, ...]
) -> None:
    stddevs = sc.empty(dims=var.dims, shape=var.shape, dtype=var.dtype, unit=var.unit)
    try:
        dataset.read_direct(stddevs.values, source_sel=index)
    except TypeError:
        stddevs.values = dataset[index].squeeze()
    # According to the standard, errors must have the same shape as the data.
    # This is not the case in all files we observed, is there any harm in
    # attempting a broadcast?
    var.variances = np.broadcast_to(
        sc.pow(stddevs, sc.scalar(2)).values, shape=var.shape
    )
